#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 15 15:00:29 2019

This file handles all interactions with the hdf and image files, such as 
loading or saving a particular image.
"""
from nd2reader import ND2Reader
import numpy as np
import re
import h5py
import os.path
import skimage
import skimage.io

from ..unet import hungarian as hu
import logging
import os
logging.basicConfig(
    format='%(asctime)s %(levelname)s %(funcName)s: %(message)s',
    level=os.environ.get("LOGLEVEL", "WARNING")
)
log = logging.getLogger(__name__)


class Reader:
    
    
    def __init__(self, hdfpathname, newhdfname, nd2pathname):
        """
        Initializes the data corresponding to the sizes of the pictures,
        the number of different fields of views(Npos) taken in the experiment.
        And it also sets the number of time frames per field of view.
        """
        # Identify filetype of image file
        _, self.extension = os.path.splitext(nd2pathname)
        self.isnd2 = self.extension == '.nd2'
        self.isfolder = self.extension == ''
        self.issingle = self.extension in ['.tif','.tiff',
                                            '.jpg','.jpeg','.png','.bmp',
                                           '.pbm','.pgm','.ppm','.pxm','.pnm','.jp2',
                                           '.TIF','.TIFF',
                                            '.JPG','.JPEG','.PNG','.BMP',
                                           '.PBM','.PGM','.PPM','.PXM','.PNM','.JP2']
        
        self.nd2path = nd2pathname # path name is nd2path for legacy reasons
        self.hdfpath = hdfpathname
        self.newhdfpath = newhdfname
        
        if self.isnd2:
            with ND2Reader(self.nd2path) as images:
                self.sizex = images.sizes['x']
                self.sizey = images.sizes['y']
                self.sizet = images.sizes['t']
                try:
                    self.sizec = images.sizes['c']
                except KeyError:
                    self.sizec = 1
                try:
                    self.Npos  = images.sizes['v']
                except KeyError:
                    self.Npos  = 1
                self.channel_names = images.metadata['channels']
                
        elif self.issingle:
#            with pytiff.Tiff(self.nd2path) as handle:
#                self.sizey, self.sizex = handle.shape #SJR: changed by me
#                self.sizec = 1
#                self.sizet = handle.number_of_pages
#                self.Npos = 1
#                self.channel_names = ['Channel1']
            im = skimage.io.imread(self.nd2path)

            if im.ndim==3:
                # num pages should be smaller than x or y dimension, very unlikely not to be the case
                if im.shape[2] < im.shape[0] and im.shape[2] < im.shape[1]:  
                    im = np.moveaxis(im, -1, 0) # move last axis to first
                self.sizet, self.sizey, self.sizex = im.shape
            else:
                self.sizey, self.sizex = im.shape
                self.sizet = 1
                
            self.Npos = 1
            self.channel_names = ['Channel1']
                
        elif self.isfolder:
            filelist = sorted(os.listdir(self.nd2path))  
            # filter filelist for supported image files
            filelist = [f for f in filelist if re.search(r".png|.tif|.jpg|.bmp|.jpeg|.pbm|.pgm|.ppm|.pxm|.pnm|.jp2"
                                                         "|.PNG|.TIF|.JPG|.BMP|.JPEG|.PBM|.PGM|.PPM|.PXM|.PNM|.JP2", f)]          
            for f in filelist:
                if f.startswith('.'):
                    filelist.remove(f)
            self.sizey = 0
            self.sizex = 0
            self.sizec = 1
            self.Npos = 1
            self.sizet = len(filelist)
            
            for f in filelist:
                im = skimage.io.imread(os.path.join(self.nd2path , f))
                if im.ndim==3:
                    self.sizec = max(self.sizec, im.shape[0])
                    self.sizey = max(self.sizey, im.shape[1]) #SJR: changed by me
                    self.sizex = max(self.sizex, im.shape[2]) #SJR: changed by me
                else:
                    self.sizey = max(self.sizey, im.shape[0]) #SJR: changed by me
                    self.sizex = max(self.sizex, im.shape[1]) #SJR: changed by me

            self.channel_names = [f'Channel{i}' for i in range(1,self.sizec+1)]

        #create the labels which index the masks with respect to time and 
        #fov indices in the hdf5 file
        self.fovlabels = []
        self.tlabels = []        
        self.InitLabels()
        
        self.default_channel = 0
        self.name = self.hdfpath
                            
        # create an new hfd5 file if no one existing already
        self.Inithdf()

        
    def InitLabels(self):
        """Create two lists containing all the possible fields of view and time
        labels, in order to access the arrays in the hdf5 file.
        """
        for i in range(0, self.Npos):
            self.fovlabels.append('FOV' + str(i))
        
        for j in range(0, self.sizet):
            self.tlabels.append('T'+ str(j))
    
    
    def Inithdf(self):
        """If the file already exists then it is loaded else 
        a new hdf5 file is created and for every fields of view
        a new group is created in the createhdf method
        """
        newFileExists = os.path.isfile(self.newhdfpath)
        if not self.hdfpath and not newFileExists:            
            return self.Createhdf()
             
        else:
            if not self.hdfpath:
                self.hdfpath = self.newhdfpath
            filenamewithpath, extension = os.path.splitext(self.hdfpath)
                
            # If mask file is a tiff file
            if (extension == '.tiff' or extension == '.tif' or 
               extension == '.TIFF' or extension == '.TIF'):
                im = skimage.io.imread(self.hdfpath)
                imdims = im.shape
                
                # num pages should be smaller than x or y dimension, very unlikely not to be the case
                if len(imdims) == 3 and imdims[2] < imdims[0] and imdims[2] < imdims[1]:  
                    im = np.moveaxis(im, -1, 0) # move last axis to first

                with h5py.File(filenamewithpath + '.h5', 'w') as hf:
                    hf.create_group('FOV0')  
        
                    if im.ndim==2:
                        hf.create_dataset('/FOV0/T0', data = im, compression = 'gzip')
                    elif im.ndim==3:
                        nim = im.shape[0]
                        for i in range(nim):
                            hf.create_dataset('/FOV0/T{}'.format(i), 
                                              data = im[i,:,:], compression = 'gzip')

                self.hdfpath = filenamewithpath + '.h5'
            
            
    def Createhdf(self):
        """In this method, for each field of view one group is created. And 
        in each one of these group, there will be for each time frame a 
        corresponding dataset equivalent to a 2d array containing the 
        corresponding masks data (segmented/thresholded/predicted).
        """
                    
        self.hdfpath = self.newhdfpath
        
        hf = h5py.File(self.hdfpath, 'w')
        for i in range(0, self.Npos):
            grpname = self.fovlabels[i]
            hf.create_group(grpname)
        hf.close()


    def LoadMask(self, currentT, currentFOV):
        """this method is called when one mask should be loaded from the file 
        on the disk to the user's buffer. If there is no mask corresponding
        in the file, it creates the mask corresponding to the given time and 
        field of view index and returns an array filled with zeros.
        """
        
        file = h5py.File(self.hdfpath,'r+')
        if self.TestTimeExist(currentT,currentFOV,file):
            mask = np.array(file['/{}/{}'.format(self.fovlabels[currentFOV], self.tlabels[currentT])], dtype = np.uint16)
            log.debug('load mask')
            file.close()
            
            return mask
        
        else:
            zeroarray = np.zeros([self.sizey, self.sizex],dtype = np.uint16)
            file.create_dataset('/{}/{}'.format(self.fovlabels[currentFOV], self.tlabels[currentT]), 
                                data = zeroarray, compression = 'gzip')
            log.debug('create dataset with zeroarray')
            file.close()
            return zeroarray
            
            
    def TestTimeExist(self, currentT, currentFOV, file=None):
        """This method tests if the array which is requested by LoadMask
        already exists or not in the hdf file.
        
        If file is None, then it opens the h5py File. Otherwise allows to pass
        an already open file. 
        """
        def closefile(f):
            if file is None:
                f.close()
                
        def openfile(file):
            if file is None:
                return h5py.File(self.hdfpath, 'r+')
            else:
                return file
                            
        if currentT <= len(self.tlabels) - 1 and currentT >= 0:
            f = openfile(file)
            for t in f['/{}'.format(self.fovlabels[currentFOV])].keys():
                # currentT is a number
                # self.tlabels is some string that indexes the time point? E.g., T0?
                if t == self.tlabels[currentT]:
                    closefile(f)
                    return True
            closefile(f)
            return False
        else:
            return False

            
    def SaveMask(self, currentT, currentFOV, mask):
        """This function is called when the user wants to save the mask in the
        hdf5 file on the disk. It overwrites the existing array with the new 
        one given in argument. 
        If it is a new mask, there should already
        be an existing null array which has been created by the LoadMask method
        when the new array has been loaded/created in the main before calling
        this save method.
        """
        
        file = h5py.File(self.hdfpath, 'r+')
        
        if self.TestTimeExist(currentT,currentFOV,file):
            dataset= file['/{}/{}'.format(self.fovlabels[currentFOV], self.tlabels[currentT])]
            dataset[:] = mask
            log.debug('save mask for FOV {} and frame {} to file'.format(self.fovlabels[currentFOV], self.tlabels[currentT]))
            file.close()
            
        else:
            file.create_dataset('/{}/{}'.format(self.fovlabels[currentFOV], self.tlabels[currentT]), data = mask, compression = 'gzip')
            log.debug('create dateset and save mask to file')
            file.close()
        
        
    def TestIndexRange(self,currentT, currentfov):
        """this method receives the time and the fov index and checks
        if it is present in the images data.
        """
        
        if currentT < (self.sizet-1) and currentfov < self.Npos:
            return True
        if currentT == self.sizet - 1 and currentfov < self.Npos:
            return False


    def LoadOneImage(self,currentT, currentfov, current_channel=None):
        """This method returns from the nd2 file, the picture requested by the 
        main program as an array. It fixes the fov index and iterates over the 
        time index.
        """
        if not (currentT < self.sizet and currentfov < self.Npos):
            return None
        
        if self.isnd2:
            with ND2Reader(self.nd2path) as images:
                try:
                    images.default_coords['v'] = currentfov
                except ValueError:
                    pass
                try:
                    images.default_coords['c'] = current_channel if current_channel is not None else self.default_channel
                except ValueError:
                    pass
                images.iter_axes = 't'
                im = images[currentT]
                outputarray = np.array(im, dtype = np.uint16)
                return outputarray

                
        elif self.issingle:
            full = skimage.io.imread(self.nd2path)
            if full.ndim==2:
                im = full
            elif full.ndim==3:
                # num pages should be smaller than x or y dimension, very unlikely not to be the case
                if full.shape[2] < full.shape[0] and full.shape[2] < full.shape[1]:  
                    full = np.moveaxis(full, -1, 0) # move last axis to first
                im = full[currentT]

            outputarray = np.array(im, dtype = np.uint16)
            return outputarray

                                
        elif self.isfolder:
            filelist = sorted(os.listdir(self.nd2path))
            # filter filelist for supported image files
            filelist = [f for f in filelist if re.search(r".png|.tif|.jpg|.bmp|.jpeg|.pbm|.pgm|.ppm|.pxm|.pnm|.jp2"
                                                         "|.PNG|.TIF|.JPG|.BMP|.JPEG|.PBM|.PGM|.PPM|.PXM|.PNM|.JP2", f)]
            for f in filelist:
                if f.startswith('.'):
                    filelist.remove(f)
            
            im = skimage.io.imread(os.path.join(self.nd2path , filelist[currentT]))
            if im.ndim==2:
                im = np.pad(im,( (0, self.sizey - im.shape[0]) , (0, self.sizex -  im.shape[1] ) ),constant_values=0) # pad with zeros so all images in the same folder have same size
                outputarray = np.array(im, dtype = np.uint16)
            elif im.ndim ==3:
                im = np.pad(im,( (0, self.sizec - im.shape[0]) , (0, self.sizey - im.shape[1]) , (0, self.sizex -  im.shape[2] ) ),constant_values=0)
                # number of channels should be smaller than x and y
                if im.shape[2] < im.shape[0] and im.shape[2] < im.shape[1]:  
                    im = np.moveaxis(im, -1, 0) # move last axis to first
                if current_channel == None:
                    im = im[self.default_channel]
                else:
                    im = im[current_channel]
                outputarray = np.array(im, dtype = np.uint16)
            else:
                outputarray = np.zeros([self.sizey, self.sizex],dtype = np.uint16)
                print("Error: image has wrong number of dimensions")
            return outputarray              

    
    def LoadImageChannel(self,currentT, currentFOV, ch):
        """Loads image at specified time, FOV and channel. Only for nd2 files"""
        if self.isnd2:
            with ND2Reader(self.nd2path) as images:
                try:
                    images.default_coords['v'] = currentFOV
                except ValueError:
                    pass
                try:
                    images.iter_axes = 'c'
                except ValueError:
                    pass
                images.default_coords['t'] = currentT
                im = images[ch]
                return np.array(im)
        
        elif self.issingle:
            return self.LoadOneImage(currentT, currentFOV, current_channel=ch)
                
        elif self.isfolder:
            return self.LoadOneImage(currentT, currentFOV,current_channel=ch)


    def CellCorrespondence(self, currentT, currentFOV):
        """Performs tracking, handles loading of the images. If the image to 
        track has no precedent, returns unaltered mask. If no mask exists
        for the current timeframe, returns zero array."""
        filemasks = h5py.File(self.hdfpath, 'r+')
        log.debug('Reader.CellCorrespondence')
        
        if self.TestTimeExist(currentT-1, currentFOV, filemasks):
            prevmask = np.array(filemasks['/{}/{}'.format(self.fovlabels[currentFOV], 
                                                          self.tlabels[currentT-1])])
            # A mask exists for both time frames
            if self.TestTimeExist(currentT, currentFOV, filemasks):
                nextmask = np.array(filemasks['/{}/{}'.format(self.fovlabels[currentFOV],
                                                              self.tlabels[currentT])])             
                newmask = hu.correspondence(prevmask, nextmask)
                out = newmask
                log.debug('make new mask')
            # No mask exists for the current timeframe, return empty array
            else:
                null = np.zeros([self.sizey, self.sizex])
                log.warn('No mask exists in FOV {} for the current timeframe {}, return empty array'.format(self.fovlabels[currentFOV],self.tlabels[currentT-1]))
                out = null
        
        else:
            # Current mask exists, but no previous - returns current mask unchanged
            if self.TestTimeExist(currentT, currentFOV, filemasks):
                nextmask = np.array(filemasks['/{}/{}'.format(self.fovlabels[currentFOV],
                                                              self.tlabels[currentT])]) 
                out = nextmask
                log.warn('NCurrent mask exists, but no previous - returns current mask unchanged. FOV {} and Time {}'.format(self.fovlabels[currentFOV],self.tlabels[currentT-1]))
            # Neither current nor previous mask exists - return empty array
            else:
                log.warn('Neither current nor previous mask exists - return empty array. FOV {} and Time {}'.format(self.fovlabels[currentFOV],self.tlabels[currentT-1]))
                null = np.zeros([self.sizey, self.sizex])
                out = null
                    
        filemasks.close()
        return out
                
                
                
        



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 17:38:58 2019
"""

from PyQt6.QtWidgets import (QDialog, QDialogButtonBox, QLineEdit, QFormLayout, 
                             QLabel, QListWidget, QAbstractItemView, QCheckBox,
                             QButtonGroup, QRadioButton, QComboBox)
from PyQt6 import QtGui


class CustomDialog(QDialog):
    def __init__(self, *args, **kwargs):
        super(CustomDialog, self).__init__(*args, **kwargs)
        
        app, = args
        maxtimeindex = app.reader.sizet
        
        self.setWindowTitle("Launch NN")
        self.setGeometry(100,100, 500,200)
        
        self.entry1 = QLineEdit()
        self.entry1.setValidator(QtGui.QIntValidator(0,int(maxtimeindex-1)))
        self.entry2 = QLineEdit()
        self.entry2.setValidator(QtGui.QIntValidator(0,int(maxtimeindex-1)))
        
        # FOV dialog
        self.listfov = QListWidget()
        self.listfov.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
        for f in range(0, app.reader.Npos):
            self.listfov.addItem('Field of View {}'.format(f+1))

        self.labeltime = QLabel("Enter range of frames ({}-{}) to segment".format(0, app.reader.sizet-1))
        
        self.entry_threshold = QLineEdit()
        self.entry_threshold.setValidator(QtGui.QDoubleValidator())
        self.entry_threshold.setText('0.5')
        
        self.entry_segmentation = QLineEdit()
        self.entry_segmentation.setValidator(QtGui.QIntValidator())
        self.entry_segmentation.setText('5')
                
        flo = QFormLayout()
        flo.addWidget(self.labeltime)
        flo.addRow('Start from frame:', self.entry1)
        flo.addRow('End at frame:', self.entry2)        
        flo.addRow('Select field(s) of view:', self.listfov)
        flo.addRow('Threshold value:', self.entry_threshold)
        flo.addRow('Min. distance between seeds:', self.entry_segmentation)
        
        self.mic_type = QComboBox()
        self.mic_type.addItem("Image type", None)
        self.mic_type.addItem("Bright-field budding yeast", "bf")
        self.mic_type.addItem("Phase contrast budding yeast", "pc")
        self.mic_type.addItem("Bud phase contrast fission", "fission")

        self.mic_type.setCurrentIndex(0)
        flo.addRow("Select image type: ", self.mic_type)
        
        
        self.device_selection = QComboBox()
        self.device_selection.addItem("CPU", "cpu")
        self.device_selection.addItem("GPU", "cuda")
        self.device_selection.setCurrentIndex(0)
        flo.addRow("Select device for running neural network: ", self.device_selection)
        
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        flo.addWidget(self.buttonBox)
        self.setLayout(flo)
        

        



from PyQt6.QtWidgets import QApplication, QDialog, QProgressBar, QVBoxLayout, QPushButton, QLabel
import time

class ProgressBar(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        # set the dialog to be a modal that blocks input to the main window
        self.setModal(True)

        # set the title of the dialog
        self.setWindowTitle("Progress")

        # create a progress bar widget
        self.progress = QProgressBar(self)

        # set the minimum and maximum values of the progress bar
        self.progress.setMinimum(0)
        self.progress.setMaximum(100)

        # create a label widget to display the progress status
        self.status = QLabel("Processing...", self)

        # create a layout to add the widgets to
        layout = QVBoxLayout(self)
        layout.addWidget(self.status)
        layout.addWidget(self.progress)
        self.setLayout(layout)

        # set initial value
        self.progress.setValue(0)
        self.status.setText("Processing...")

        # show the dialog
        self.show()

    def update_progress(self, value):
        self.progress.setValue(value)
    def set_status(self, t):
        self.status.setText("Processing... {}%".format(t))


import numpy as np
import pandas as pd
from munkres import Munkres
from sklearn.preprocessing import scale
from sklearn.metrics.pairwise import euclidean_distances


def correspondence(prev, curr):
    """
    Corrects correspondence between previous and current mask, returns current
    mask with corrected cell values. New cells are given the unique identifier
    starting at max(prev)+1. 
    
    This is done by embedding every cell into a feature space consisting of
    the center of mass and the area. The pairwise euclidean distance is 
    calculated between the cells of the previous and current frame. This is 
    then used as a cost for the bipartite matching problem which is in turn
    solved by the Hungarian algorithm as implemented in the munkres package.
    """
    newcell = np.max(prev) + 1
    
    hu_dict = hungarian_align(prev, curr)
    new = curr.copy()
    for key, val in hu_dict.items():
        # If new cell
        if val == -1:
            val = newcell
            newcell += 1
        
        new[curr==key] = val
        
    return new


def hungarian_align(m1, m2):
    """
    Aligns the cells using the hungarian algorithm using the euclidean distance as 
    cost. 
    Returns dictionary of cells in m2 to cells in m1. If a cell is new, the dictionary 
    value is -1.
    """
    dist, ix1, ix2 = cell_distance(m1, m2)
    
    # If dist couldn't be calculated, return dictionary from cells to themselves 
    if dist is None:
        unique_m2 = np.unique(m2)
        return dict(zip(unique_m2, unique_m2))
    
    solver = Munkres()
    indexes = solver.compute(make_square(dist))
    
    # Create dictionary of cell indicies
    d = dict([(ix2.get(i2, -1), ix1.get(i1, -1)) for i1, i2 in indexes])
    d.pop(-1, None)  
    return d


def cell_to_features(im, c, nsamples=None, time=None):
    """Embeds cell c in image im into feature space"""
    coord = np.argwhere(im==c)
    area = coord.shape[0]
    
    if nsamples is not None:
        samples = np.random.choice(area, min(nsamples, area), replace=False)
        sampled = coord[samples,:]
    else:
        sampled = coord
    
    com = sampled.mean(axis=0)
    
    return {'cell': c,
            'time': time,
            'sqrtarea': np.sqrt(area),
            'area': area,
            'com_x': com[0],
            'com_y': com[1]}
    
    
def cell_distance(m1, m2, weight_com=3):
    """
    Gives distance matrix between cells in first and second frame, by embedding
    all cells into the feature space. Currently uses center of mass and area
    as features, with center of mass weighted with factor weight_com (to 
    make it more important).
    """
    # Modify to compute use more computed features
    #cols = ['com_x', 'com_y', 'roundness', 'sqrtarea']
    cols = ['com_x', 'com_y', 'area']

    def get_features(m, t):
        cells = list(np.unique(m))
        if 0 in cells:
            cells.remove(0)
        features = [cell_to_features(m, c, time=t) for c in cells]
        return pd.DataFrame(features), dict(enumerate(cells))
    
    # Create df, rescale
    feat1, ix_to_cell1 = get_features(m1, 1)
    feat2, ix_to_cell2 = get_features(m2, 2)
    
    # Check if one of matrices doesn't contain cells
    if len(feat1)==0 or len(feat2)==0:
        return None, None, None
    
    df = pd.concat((feat1, feat2))
    df[cols] = scale(df[cols])
    
    # give more importance to center of mass
    df[['com_x', 'com_y']] = df[['com_x', 'com_y']] * weight_com

    # pairwise euclidean dist
    dist = euclidean_distances(
        df.loc[df['time']==1][cols],
        df.loc[df['time']==2][cols]
    )
    return dist, ix_to_cell1, ix_to_cell2
    
    
def zero_pad(m, shape):
    """Pads matrix with zeros to be of desired shape"""
    out = np.zeros(shape)
    nrow, ncol = m.shape
    out[0:nrow, 0:ncol] = m
    return out


def make_square(m):
    """Turns matrix into square matrix, as required by Munkres algorithm"""
    r,c = m.shape
    if r==c:
        return m
    elif r>c:
        return zero_pad(m, (r,r))
    else:
        return zero_pad(m, (c,c))

    



# Run:
#
# python Launch_NN_command_line.py -i DIRECTORY/IMAGE_FILE -m OUTPUT_MASK_FILE --path_to_weights PATH_TO_HDF5_FILE --fov N --range_of_frames n1 n2 --min_seed_dist 5 --threshold 0.5
# 
# or:
#
# python Launch_NN_command_line.py -i DIRECTORY/IMAGE_FILE -m OUTPUT_MASK_FILE --image_type pc_OR_bf               --fov N --range_of_frames n1 n2 --min_seed_dist 5 --threshold 0.5

import time
import tqdm

import sys

#Import all the other python files
#this file handles the interaction with the disk, so loading/saving images
#and masks and it also runs the neural network.


from yeaz.unet.segment import segment
from yeaz.disk import Reader as nd
import argparse
import skimage
from yeaz.unet import neural_network as nn

import torch

def LaunchPrediction(im, mic_type, pretrained_weights=None, device='cpu'):
    """It launches the neural neutwork on the current image and creates 
    an hdf file with the prediction for the time T and corresponding FOV. 
    """
    im = skimage.exposure.equalize_adapthist(im)
    im = im*1.0;	
    pred = nn.prediction(im, mic_type, pretrained_weights, device=device)
    return pred



def ThresholdPred(thvalue, pred):
    """Thresholds prediction with value"""
    if thvalue == None:
        thresholdedmask = nn.threshold(pred)
    else:
        thresholdedmask = nn.threshold(pred, thvalue)
    return thresholdedmask

def LaunchInstanceSegmentation(reader, image_type, fov_indices=[0], time_value1=0, time_value2=0, thr_val=None, min_seed_dist=5, path_to_weights=None, device='cpu'):
    if (device == 'cuda') and torch.cuda.is_available():
        f_device = 'cuda'
    else:
        f_device = 'cpu' 
    # cannot have both path_to_weights and image_type supplied
    if (image_type is not None) and (path_to_weights is not None):
        print("image_type and path_to_weights cannot be both supplied.")
        return
    

    # check if correct imaging value
    if (image_type not in ['bf', 'pc']) and (path_to_weights is None):
        print("Wrong imaging type value ('{}')!".format(image_type),
              "imaging type must be either 'bf' or 'pc'")
        return

    # check range_of_frames constraint
    if time_value1 > time_value2 :
        print("Error", 'Invalid Time Constraints')
        return
    
    # displays that the neural network is running
    print('Running the neural network on {} ...'.format(f_device))
    
    for fov_ind in tqdm.tqdm(fov_indices, desc='FOV', position=0):

        #iterates over the time indices in the range
        for t in tqdm.tqdm(range(time_value1, time_value2+1), desc='Time', position=1, leave=False):         
            # print('--------- Segmenting field of view:',fov_ind,'Time point:',t)

            #calls the neural network for time t and selected fov
            im = reader.LoadOneImage(t, fov_ind)

            try:
                pred = LaunchPrediction(im, image_type, pretrained_weights=path_to_weights, device=f_device)
            except ValueError:
                print('Error! ',
                      'The neural network weight files could not '
                      'be found. \nMake sure to download them from '
                      'the link in the readme and put them into '
                      'the folder unet, or specify a path to a custom weights file with -w argument.')
                return

            thresh = ThresholdPred(thr_val, pred)
            seg = segment(thresh, pred, min_seed_dist)
            reader.SaveMask(t, fov_ind, seg)
            print('--------- Finished segmenting.')
            
            # apply tracker if wanted and if not at first time
            temp_mask = reader.CellCorrespondence(t, fov_ind)
            reader.SaveMask(t, fov_ind, temp_mask)

def main(args):

    if '.h5' in args.mask_path:
        args.mask_path = args.mask_path.replace('.h5','')

    reader = nd.Reader("", args.mask_path+'.h5', args.image_path)

    LaunchInstanceSegmentation(reader, args.image_type, args.fov,
                               args.range_of_frames[0],  args.range_of_frames[1], 
                               args.threshold, args.min_seed_dist, 
                               args.path_to_weights, device=args.device)

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-i', '--image_path', type=str, help="Specify the path to a single image or to a folder of images", required=True)
    parser.add_argument('-m', '--mask_path', type=str, help="Specify where to save predicted masks", required=True)
    parser.add_argument('--image_type', type=str, help="Specify the imaging type, possible types are 'bf', 'pc', and 'fission'. Supersedes path_to_weights.")
    parser.add_argument('--path_to_weights', default=None, type=str, help="Specify weights path.")
    parser.add_argument('--fov', default=[0], nargs='+', type=int, help="Specify field of view index (can specify more than one with space between them).")
    parser.add_argument('--range_of_frames', nargs=2, default=[0,0], type=int, help="Specify start and end in range of frames. (e.g. 0 10)")
    parser.add_argument('--threshold', default=None, type=float, help="Specify threshold value.")
    parser.add_argument('--min_seed_dist', default=5, type=int, help="Specify minimum distance between seeds.")
    parser.add_argument('--device', default='cpu', type=str, help="Specify device to run on (cpu or cuda).")
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This script is the main script used to produce a GUI which should help for
cell segmentation. This script can only read .nd2 files containing the 
images of cells, especially it displays for each recorded positions (field
of view) the pictures in the time axis.

The script opens first a window which allows you to load an nd2 file and
to load or create an hdf file. The hdf file contains all the masks, so if
it is the first the user segments an nd2 file, a new one should be created.
And it can be then loaded for later use.  Along with new hdf file, created
by the name entered by the user (say filename), it creates three other hdf 
files (filename_predicted.h5, filename_thresholded.h5 and 
filename_segmented.h5) these contain all the steps of the NN to get to the 
segmented picture. 

After the first window is finished a second one opens, where at each time 
index, three pictures 
are displayed the t-1 picture, the t picture (current frame which can be
edited) and the t+1 picture. Using the arrows one can navigate through time.
On top of the picture, there is always a mask which is displayed, if no cells
are present in the mask then the mask is blank and the user does not see it. 
If one wants to hand anmotate the pictures, one can just start to draw on the
picture using the different functions (New Cell, Add Region, Brush, Eraser,
Save Mask, ...) and the informations will be saved in the mask overlayed on 
top of the pictures. 

If one wants to segment using a neural network, one can press the
corresponding button (Launch CNN) and select the time range and 
the field of views on which the neural network is applied.

Once the neural network has finished predicting, there are still no visible
masks, but on the field of views and time indices where the NN has been
applied, the threshold and segment buttons are enabled. By checking these
two buttons one can either display the thresholded image of the prediction or
display the segmentation of the thresholded prediction.

At this stage, one can correct the segmentation of the prediction using
the functions (New Cell, Add Region, etc..) by selecting the Segment
checkbox and then save them using the Save Seg button.
If the user is happy with the segmentation, the Cell Correspondence button 
can be clicked. Until then, the cells get random numbers attributed by
the segmentation algorithm. In order to keep track of the cell through time,
the same cells should have the same number between two different time pictures.
This can be (with some errors) achieved by the Cell Correspondence button,
which tries to attribute the same number to corresponding cells in time.
After that, the final mask is saved and it is always visible when you go on
the corresponding picture. This mask can also be corrected using the 
usual buttons (because the Cell Correspondence makes also mistakes). 

"""
import sys
import numpy as np
import pandas as pd
import h5py
import skimage

# For writing excel files
#from openpyxl import load_workbook
#from openpyxl import Workbook

#Import from matplotlib to use it to display the pictures and masks.
from matplotlib.backends.qt_compat import QtWidgets
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar

from sklearn.decomposition import PCA
import imageio
from PIL import Image, ImageDraw

# tqdm for progress bar
import tqdm
import time

# Import everything for the Graphical User Interface from the PyQt6 library.
from PyQt6.QtWidgets import (QApplication, QMainWindow, QDialog, QSpinBox,
    QMessageBox, QPushButton, QCheckBox, QStatusBar, QLabel)
from PyQt6 import QtGui
from PyQt6.QtGui import QAction
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPalette, QColor

#Import all the other python files
#this file handles the interaction with the disk, so loading/saving images
#and masks and it also runs the neural network.
from .disk import Reader as nd

#this file contains a dialog window that takes two integers as entry to swap
#two cell values
from .misc import ExchangeCellValues as ecv

#this file contains a window that opens to change the value of one cell. It 
#is opened as soon as the user presses with the left click on a specific cell.
from .misc import ChangeOneCellValue as cocv

#this file contains a dialog window where a time range and the field of views
#can be selected to then launch a prediction of the neural network on
#a specific range of pictures.
from .unet import LaunchBatchPrediction as lbp

#this file contains a dialog window where a time range can be selected to retrack
from .misc import BatchRetrack as br

#this file initializes all the buttons present in the gui, sets the shortcuts
#to these buttons and also connect the buttons to the function that are 
#triggered when the buttons are pressed.
from .init import InitButtons

#this file contains the layout of the main window so it justs puts the buttons
#and the pictures at the desired position in the main window.
from .init import InitLayout

# PlotCanvas for fast plotting
from .misc.PlotCanvas import PlotCanvas

from .misc import Extract as extr

from .disk.image_loader import load_image
from .unet.segment import segment
from .unet import neural_network as nn
from .misc.ProgressBar import ProgressBar

import warnings
# warnings.filterwarnings('ignore')

import logging
import os

# Configure the root logger
logging.basicConfig(
    format='%(asctime)s %(levelname)s %(funcName)s: %(message)s',
    level=os.environ.get("LOGLEVEL", "WARNING"),
    filename='yeaz_GUI.log'
)
log = logging.getLogger(__name__)


if getattr(sys, 'frozen', False):
    path_icons = os.path.join(sys._MEIPASS, "icons/")
    path_weights  = os.path.join(sys._MEIPASS, 'unet/')
    
else:
    path_icons = './icons/'
    path_weights = './unet/'


class NavigationToolbar(NavigationToolbar):
    """This is the standard matplotlib toolbar but only the buttons
    that are of interest for this gui are loaded. These buttons allow 
    to zoom into the pictures/masks and to navigate in the zoomed picture. 
    A Home button can be used to set the view back to the original view.
    """
    toolitems = [t for t in NavigationToolbar.toolitems if
                 t[0] in ('Home', 'Pan', 'Zoom','Back', 'Forward')]
      
class App(QMainWindow):
    """This class creates the main window.
    """

    def __init__(self, nd2pathstr, hdfpathstr, newhdfstr):
        super().__init__()
        self.setWindowTitle('YeaZ')
        log.info('initiating application')

        # all these ids are integers which are used to set a connection between
        # the button and the function that this button calls.
        # There are three of them because it happens that one can trigger three 
        # different functions with one button.        
        self.id = 0
        self.id2 = 0
        self.id3 = 0 

        self.reader = nd.Reader(hdfpathstr, newhdfstr, nd2pathstr)
        
        # these variables are used to create/read/load the excel file used
        # to write the fluorescence values extracted. For each field of view,
        # the user will be asked each time to create a new xls file for the 
        # field of view or to load an existing field of view (this is the role
        # of the boolean variable)
        self.xlsfilename = ''
        self.nd2path = nd2pathstr
        
        # Set the indices for the time axis and the field of view index. These
        # indices represent everywhere the current picture (the one that can be
        # edited, i.e. the time t frame)
        self.Tindex = 0
        self.FOVindex = 0
        
        # loading the first images of the cells from the nd2 file
        self.currentframe = self.reader.LoadOneImage(self.Tindex,self.FOVindex)
        
        # check if the t+1 time frame exists, avoid failure if there is only
        # one picture in the folder/nd2 file
        if self.Tindex+1 < self.reader.sizet:
            self.nextframe = self.reader.LoadOneImage(self.Tindex+1, self.FOVindex)
        else:
            self.nextframe = np.zeros([self.reader.sizey, self.reader.sizex])
        
        self.previousframe = np.zeros([self.reader.sizey, self.reader.sizex])

        # loading the first masks from the hdf5 file
        self.mask_curr = self.reader.LoadMask(self.Tindex, self.FOVindex)
        self.mask_previous = np.zeros([self.reader.sizey, self.reader.sizex])
        
        # check if the t+1 mask exists, avoid failure if there is only
        # one mask in the hdf file
        if self.Tindex+1 < self.reader.sizet:
            self.mask_next = self.reader.LoadMask(self.Tindex+1, self.FOVindex)
        else:
            self.mask_next = np.zeros([self.reader.sizey, self.reader.sizex])
        
        # creates a list of all the buttons, which will then be used in order
        # to disable all the other buttons at once when one button/function
        # is pressed/used in the gui.
        self.buttonlist = []
        
        # setting buttons as attributes
        # the shortcuts for the buttons, the functions to which they are
        # connected to,... are all set up in the ButtonInit file which is called
        # in the self.initUI() method below.
        self.button_newcell = QPushButton("New cell")
        self.buttonlist.append(self.button_newcell)
        
        self.button_add_region = QPushButton("Add region")
        self.buttonlist.append(self.button_add_region)
        
        self.button_drawmouse = QPushButton('Brush')
        self.buttonlist.append(self.button_drawmouse)
        
        self.button_eraser = QPushButton('Eraser')
        self.buttonlist.append(self.button_eraser)
        
        self.label_brushsize = QLabel('Brush/Eraser radius:')
        self.spinbox_brushsize = QSpinBox()
        self.buttonlist.append(self.spinbox_brushsize)
        
        self.button_exval = QPushButton('Exchange cell IDs')
        self.buttonlist.append(self.button_exval)
        
        self.button_showval = QCheckBox('Show cell IDs')
        self.buttonlist.append(self.button_showval)
        
        self.button_hidemask = QCheckBox('Hide mask')
        self.buttonlist.append(self.button_hidemask)
        
        self.button_split = QPushButton('Split cell')
        self.buttonlist.append(self.button_split)
        
        self.button_mergewithneighbors = QPushButton('Merge cells')
        self.buttonlist.append(self.button_mergewithneighbors)
        
        self.button_nextframe = QPushButton("Next time frame")
        self.buttonlist.append(self.button_nextframe)
        
        self.button_previousframe = QPushButton("Previous time frame")
        self.buttonlist.append(self.button_previousframe)
        
        self.button_cnn = QPushButton('Launch CNN')
        self.buttonlist.append(self.button_cnn)
        
        self.button_cellcorrespondence = QPushButton('Retrack')
        self.buttonlist.append(self.button_cellcorrespondence)
        
        self.button_changecellvalue = QPushButton('Change cell ID')
        self.buttonlist.append(self.button_changecellvalue)        
        
        self.button_extractfluorescence = QPushButton('Extract')
        self.buttonlist.append(self.button_extractfluorescence)
        
        self.button_hide_show = QPushButton('CNN')
        self.buttonlist.append(self.button_hide_show)
        
        self.initUI()


    def initUI(self):
        """Initializing the widgets contained in the window. 
        Especially, it creates the widget to plot the 
        pictures/masks by creating an object of the PlotCanvas class self.m. 
        Every interaction with the masks or the pictures (loading new
        frames/editing the frames/masks) occurs through this class.
        
        This method initializes all the buttons with the InitButtons file.
        It connects the buttons to the functions that they should trigger,
        it sets the shortcuts to the buttons, a tool tip, 
        eventually a message on the status bar when the user hovers 
        over the button, etc..
        
        This function also sets all the layout in the InitLayout file. It 
        takes and places the widgets (buttons, canvas, toolbar).

        The function initializes a Menu Bar to have a menu which can be 
        improved later on.
        It sets a toolbar of the matplotlib library and hides it. But it allows
        to connect to the functions of this toolbar through "homemade" 
        QPushButtons instead of the ones provided by matplotlib.
        Finally, it sets a StatusBar which displays some text to describe
        the use of some buttons, or to show that the program is working on 
        something (running the neural network, loading frames, etc...)
        
        After all this has been initialized, the program is ready to be used.
        """
        log.info('initiate UI')
        self._main = QtWidgets.QWidget()
        self.setCentralWidget(self._main)

        # Here our canvas is created where using matplotlib, 
        # one can plot data to display the pictures and masks.
        self.m = PlotCanvas(self)
        
        # Initialize all the buttons that are needed and the functions that are 
        # connected when the buttons are triggered.
        InitButtons.Init(self)
        InitLayout.Init(self)
        
        # MENU, TOOLBAR AND STATUS BAR
        # creates a menu just in case, some other functions can be added later
        # in this menu.
#        menubar = self.menuBar()
#        self.fileMenu = menubar.addMenu('File')   
#        self.saveactionmenu = QAction('Save')
#        self.fileMenu.addAction(self.saveactionmenu)
#        self.saveactionmenu.triggered.connect(self.SaveMask)
        
        # hide the toolbar and instead of the original buttons of matplotlib,
        # QPushbuttons are used and are connected to the functions of the toolbar
        # it is than easier to interact with these buttons (for example to 
        # to disable them and so on..)
        self.Nvgtlbar = NavigationToolbar(self.m, self)
        self.addToolBar(self.Nvgtlbar)
        self.Nvgtlbar.hide()
        
        # creates a status bar with user instructions
        self.statusBar = QStatusBar()
        self.setStatusBar(self.statusBar)
        self.statusBarText = QLabel()
        self.statusBar.addWidget(self.statusBarText)
                
        self.show()
                
        
# -----------------------------------------------------------------------------
# FUNCTIONS LINKED TO NAVIGATION
# connect the functions of the toolbar to our custom QPushbuttons.
    def ZoomTlbar(self):
        """The button_zoom is connected to the zoom function of the toolbar 
        already present in the matplotlib library.
        
        Depending on the buttons that are active or checked, when the zoom 
        function is used, it does not disable all the buttons.
        
        If the segment and threshold button are not checked or used
        when the zoom button is clicked, it disables all the button
        using self.Disable which disables everything except the button passed
        in argument (in this case button_zoom).
        
        If the zoom button is used while the segment button is checked,
        it disables all the buttons (1st elif) except the segment button
        but once it is finished (so the zoom button becomes unchecked) 
        then it enables only the editing buttons (as long as the segment
        button is still checked) such as New Cell, Add Region, Eraser, 
        Brush,etc.. and the other toolbar buttons (3rd elif)
        
        
        If the zoom button is clicked while the threshold button is checked,
        it disables all the button except the threshold button (2nd elif).
        Once the zoom button is unchecked, it enables the toolbar buttons
        (4th elif)
        In any other case, it just enables all the buttons again.
        """
        self.Nvgtlbar.zoom()
        
        if (self.button_zoom.isChecked()):
            self.Disable(self.button_zoom)
            
        else:
            self.Enable(self.button_zoom)
        
        
    def HomeTlbar(self):
        """
        connects the home button to the home function of the matplotlib
        toolbar. It sets the view to the original view (no zoom)
        """
        self.Nvgtlbar.home()

            
    def BackTlbar(self):
        """
        It calls the back function of the matplotlib toolbar which sets the 
        view to the previous one (if the user does several zooms/pans, 
        this button allows to go back in the "history of views")
        """
        self.Nvgtlbar.back()

        
    def ForwardTlbar(self):
        """
        It calls the forward function of the matplotlib toolbar which sets the 
        view to the next one (if the user does several zooms/pans, 
        this button allows to go forward in the "history of views"
        """
        self.Nvgtlbar.forward()

    
    def PanTlbar(self):
        """The button_pan is connected to the pan function of the toolbar 
        already present in the matplotlib library.
        
        Depending on the buttons that are active or checked, when the pan 
        function is used, it does not disable all the buttons.
        
        If the segment and threshold button are not checked or used
        when the pan button is clicked, it disables all the button
        using self.Disable which disables everything except the button passed
        in argument (in this case button_pan).
        
        If the pan button is used while the segment button is checked,
        it disables all the buttons (1st elif) except the segment button
        but once it is finished (so the zoom button becomes unchecked) 
        then it enables only the editing buttons (as long as the segment
        button is still checked) such as New Cell, Add Region, Eraser, 
        Brush,etc.. and the other toolbar buttons (3rd elif)
        
        
        If the pan button is clicked while the threshold button is checked,
        it disables all the button except the threshold button (2nd elif).
        Once the pan button is unchecked, it enables the toolbar buttons
        (4th elif)
        In any other case, it just enables all the buttons again.
        """

        self.Nvgtlbar.pan()

        if (self.button_pan.isChecked()):
            self.Disable(self.button_pan)
        else:
            self.Enable(self.button_pan)

            
    def ButtonFluo(self):
        """This function is called everytime the Extract Fluorescence button is 
        clicked (self.button_extractfluorescence). 
        """        
        self.Disable(self.button_extractfluorescence)
        self.WriteStatusBar('Extracting ...')
        
        # Get last image with mask
        for time_index in range(self.reader.sizet-1, -1, -1):            
            # Test if time has a mask
            file = h5py.File(self.reader.hdfpath, 'r+')
            time_exist = self.reader.TestTimeExist(time_index, self.FOVindex, file)
            file.close()
            
            if not time_exist:
                continue
            
            # load picture and sheet
            image = self.reader.LoadImageChannel(time_index, self.FOVindex, 
                                                 self.reader.default_channel)
            mask = self.reader.LoadMask(time_index, self.FOVindex)
            
            # Break if mask is non-empty
            if mask.sum()>0:
                break
            
            if time_index==0:
                msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Error', 'No mask found', parent=self)
                msg_box.exec()
                self.Enable(self.button_extractfluorescence)
                self.ClearStatusBar()
        
        # Launch dialog with last image
        dlg = extr.Extract(image, mask, self.reader.channel_names)
        dlg.exec()
        if dlg.exit_code == 1: # Fluorescence
            self.ExtractFluo(dlg.cells, dlg.desel_cells, dlg.outfile, dlg.file_list)
        elif dlg.exit_code == 2: # Mask
            self.ExtractMask(dlg.desel_cells, dlg.outfile)
            
        self.Enable(self.button_extractfluorescence)
        self.ClearStatusBar()


    def ExtractMask(self, desel_cells, outfile):
        """Extract the mask to the specified tiff file. Only take cells 
        specified by the cell_list"""
        
        mask_list = []
        for time_index in range(0, self.reader.sizet):
            
            # Test if time has a mask
            file = h5py.File(self.reader.hdfpath, 'r+')
            time_exist = self.reader.TestTimeExist(time_index, self.FOVindex, file)
            file.close()
            
            if not time_exist:
                continue
            
            mask = self.reader.LoadMask(time_index, self.FOVindex)
            for cell in desel_cells:
                mask[mask==cell] = 0
            mask_list.append(mask)
            
        imageio.mimwrite(outfile, np.array(mask_list, dtype=np.uint16))
                        

    def ExtractFluo(self, sel_cells, desel_cells, csv_filename, channel_list):
        """This is the function that takes as argument the filepath to the xls
        file and writes in the file.
        It iterates over the different channels (or the sheets of the file,
        each channel has one sheet.), and reads the image corresponding
        to the time, field of view and channel index. It reads the already
        existing file and makes a copy in which the data will be written in it.
        
        The first step of calculating the data is to iterate through each
        cell/segment of the mask (so each cell is a submatrix of one value
        in the matrix of the mask).
        For each of these value /cell, the area is extracted as being
        the number of pixels corresponding to this cell/value. 
        (it is known from the microscope settings how to convert
        the pixel in area).
        The total intensity is just the value of the pixel and it is added over
        all the pixels corresonding to the cell/value.
        The mean is then calculated as being the total intensity divided by
        the number of pixels (which here is equal to the area also).
        With the mean it is then possible to calculate the variance of the 
        signal for one cell/value.
        
        Then, it is checked if the value of the cell (cell number) already
        exists in the first column, if it already exists it continues to
        find the column corresponding to the time index where the values
        should be written. It sets the flag to True such that it does not
        write the cell as new one and adds it at the end of the column
        
        If the value is not found in the cell number column (new cell or
        first time writing in the file), the flag is False, thus it adds the 
        cell number at the end of the column.
        It then saves the xls file.
        
        """
        # List of cell properties
        cell_list = []

        for time_index in range(0, self.reader.sizet):
            # Test if time has a mask
            file = h5py.File(self.reader.hdfpath, 'r+')
            time_exist = self.reader.TestTimeExist(time_index, self.FOVindex, file)
            file.close()
            
            if not time_exist:
                continue
            
            mask = self.reader.LoadMask(time_index, self.FOVindex)
            
            for channel in channel_list:
                # check if channel is in list of nd2 channels
                try:
                    channel_ix = self.reader.channel_names.index(channel)
                    image = self.reader.LoadImageChannel(time_index, self.FOVindex, channel_ix)
                
                # channel is a file
                except ValueError:
                    image = load_image(channel, ix=time_index)
                    
                for val in np.unique(mask):
                    # bg is not cell
                    if val == 0:
                        continue
                    # disregard cells not in cell_list
                    if (val in desel_cells):
                        continue
                    
                    # Calculate stats
                    stats = {'Cell': val,
                             'Time': time_index,
                             'Channel': channel}
                    
                    stats = {**stats,
                             **self.cell_statistics(image, mask == val)}
                    stats['Disappeared in video'] = not (val in sel_cells)
                    cell_list.append(stats)
                    
        
        # Use Pandas to write csv
        df = pd.DataFrame(cell_list)
        df = df.sort_values(['Cell', 'Time'])
        df.to_csv(csv_filename, index=False)
                    
        self.Enable(self.button_extractfluorescence)
        self.ClearStatusBar()


    def cell_statistics(self, image, mask):
        """Calculate statistics about cells. Passing None to image will
        create dictionary to zeros, which allows to extract dictionary keys"""
        if image is not None:
            cell_vals = image[mask]
            area = mask.sum()
            tot_intensity = cell_vals.sum()
            mean = tot_intensity/area if area > 0 else 0
            var = np.var(cell_vals)
            
            # Center of mass
            y,x = mask.nonzero()
            com_x = np.mean(x)
            com_y = np.mean(y)
            
            # PCA only works for multiple points
            if area > 1:
                pca = PCA().fit(np.array([x,y]).T)
                pc1_x, pc1_y = pca.components_[0,:]
                angle = np.arctan(pc1_y / pc1_x) / (2*np.pi) * 360
                v1, v2 = pca.explained_variance_
                
                len_maj = 4*np.sqrt(v1)
                len_min = 4*np.sqrt(v2)
            else:
                angle = 0
                len_maj = 1
                len_min = 1
            
        else:
            mean = 0
            var = np.nan
            tot_intensity = 0
            com_x = np.nan
            com_y = np.nan
            angle = np.nan
            len_maj = np.nan
            len_min = np.nan
        
        return {'Area': area,
                'Mean': mean,
                'Variance': var,
                'Total Intensity': tot_intensity,
                'Center of Mass X': com_x,
                'Center of Mass Y': com_y,
                'Angle of Major Axis': angle,
                'Length Major Axis': len_maj,
                'Length Minor Axis': len_min}

# -----------------------------------------------------------------------------
# NEURAL NETWORK
    def ShowHideCNNbuttons(self):
        """hide and show the buttons corresponding to the neural network.
            this function is called by the button CNN which is hidden. But
            if activated in the InitLayout.py then you can have a button
            which hides the CNN buttons (which are now on the normal also
            hidden...).
        """
        if self.button_hide_show.isChecked():
            self.button_cnn.setVisible(True)

        else:
            self.button_cnn.setVisible(False)
            

    def LaunchBatchPrediction(self):
        """This function is called whenever the button Launch CNN is pressed.
        It allows to run the neural network over a time range and selected
        field of views.
            
        It creates a dialog window with two entries, that define the time range
        and a list where the user can select the desired fields of view.
        
        Once it reads all the value, it calls the neural network function
        inside of self.PredThreshSeg and it does the prediction of the neural
        network, thresholds this prediction and then segments it.
        """
        def reset():
            self.m.UpdatePlots()
            self.ClearStatusBar()
            self.Enable(self.button_cnn)
            self.EnableCNNButtons()
        
        
        self.WriteStatusBar('Running the neural network...')
        self.Disable(self.button_cnn)

        # creates a dialog window from the LaunchBatchPrediction.py file
        dlg = lbp.CustomDialog(self)
        
        # this if tests if the user pressed 'ok' in the dialog window
        result = dlg.exec()
        if result == QDialog.DialogCode.Accepted:
            # it tests if the user has entered some values
            # if not it ignores and returns.
            if not (dlg.entry1.text()!= '' and dlg.entry2.text() != ''):
                msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Error', "No Time Specified", parent=self)
                msg_box.exec()
                reset()
                return 
            
            # reads out the entry given by the user and converts the index
            # to integers
            time_value1 = int(dlg.entry1.text())
            time_value2 = int(dlg.entry2.text())
    
            # it tests if the first value is smaller or equal such that
            # time_value1 is the lower range of the time range
            # and time_value2 the upper boundary of the range.
            if time_value1 > time_value2 :
                msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Error', 'Invalid Time Constraints', parent=self)
                msg_box.exec()
                reset()
                return
            
            # displays that the neural network is running
            self.WriteStatusBar('Running the neural network...')
    
            #it iterates in the list of the user-selected fields 
            #of view, to return the corresponding index, the function
            #dlg.listfov.row(item) is used which gives an integer
            if len(dlg.listfov.selectedItems())==0:
                msg_box = QMessageBox(QMessageBox.Icon.Critical, "Error", "No FOV Selected", parent=self)
                msg_box.exec()
            # Check if the user have selected a mic type
            
            mic_type = dlg.mic_type.currentData()
            if dlg.mic_type.currentData() is not None:
                # User has selected an option, do something with it
                log.debug(msg='Mic type selected: {}'.format(mic_type))
            else:
                # User has not selected any option, show error message
                msg_box = QMessageBox(QMessageBox.Icon.Critical, "Error", "No Image Type Selected", parent=self)
                msg_box.exec()
                reset()
                return 
            
            # decide between cuda or cpu for running neural network
            device = None
            import torch
            if (dlg.device_selection.currentData() == 'cuda') and (torch.cuda.is_available()):
                device = 'cuda'
            else:
                device = 'cpu'
            
            for item in tqdm.tqdm(dlg.listfov.selectedItems(), desc='FOV', position=0, leave=True):
                #iterates over the time indices in the range
                for t in tqdm.tqdm(range(time_value1, time_value2+1), desc='Time', position=1, leave=False):                    
                    #calls the neural network for time t and selected
                    #fov
                    if dlg.entry_threshold.text() !=  '':
                        thr_val = float(dlg.entry_threshold.text())
                    else:
                        thr_val = None
                    if dlg.entry_segmentation.text() != '':
                        seg_val = int(dlg.entry_segmentation.text())
                    else:
                        seg_val = 10
                    
                    self.PredThreshSeg(t, dlg.listfov.row(item), thr_val, seg_val,
                                       mic_type, device=device)
                    
                    # apply tracker if wanted and if not at first time
                    temp_mask = self.reader.CellCorrespondence(t, dlg.listfov.row(item))
                    self.reader.SaveMask(t,dlg.listfov.row(item), temp_mask)
            
            self.ReloadThreeMasks()
        reset()

    
    def PredThreshSeg(self, timeindex, fovindex, thr_val, seg_val, 
                      mic_type, device=None):
        """
        This function is called in the LaunchBatchPrediction function.
        This function calls the neural network function in the
        InteractionDisk.py file and then thresholds the result
        of the prediction, saves this thresholded prediction.
        Then it segments the thresholded prediction and saves the
        segmentation. 
        """
        log.debug('--------- Segmenting field of view:',fovindex,'Time point:',timeindex)
        im = self.reader.LoadOneImage(timeindex, fovindex)
        try:
            pred = self.LaunchPrediction(im, mic_type, device=device)
        except ValueError:
            msg_box = QMessageBox(QMessageBox.Icon.Critical,'Error',
                                 'The neural network weight files could not '
                                 'be found. Make sure to download them from '
                                 'the link in the readme and put them into '
                                 'the folder unet', parent=self)
            msg_box.exec()
            
            return

        thresh = self.ThresholdPred(thr_val, pred)
        seg = segment(thresh, pred, seg_val)
        self.reader.SaveMask(timeindex, fovindex, seg)
        log.debug('--------- Finished segmenting.')
          
          
    @staticmethod
    def LaunchPrediction(im, mic_type, pretrained_weights=None, device=None):
        """It launches the neural neutwork on the current image and creates 
        an hdf file with the prediction for the time T and corresponding FOV. 
        """
        im = skimage.exposure.equalize_adapthist(im)
        im = im*1.0;	
        pred = nn.prediction(im, mic_type, pretrained_weights, device=device)
        return pred


    @staticmethod
    def ThresholdPred(thvalue, pred):
        """Thresholds prediction with value"""
        if thvalue == None:
            thresholdedmask = nn.threshold(pred)
        else:
            thresholdedmask = nn.threshold(pred, thvalue)
        return thresholdedmask

    
    def SelectChannel(self, index):
        """This function is called when the button to select different channels
        is used. From the displayed list in the button, the chosen index
        corresponnds to the same index in the list of channels from the reader.
        So, it sets the default channel with the new index (called index below)
        """
        self.reader.default_channel = index
        # update the pictures using the same function as the one used to 
        # change the fields of view.
        self.ChangeFOV()
        

    def SelectFov(self, index):
        """This function is called when the button containing the list of 
        fields od view is used.
        The index correspondds to the field of view selected in the list.
        """
        # mask is automatically saved.
        self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)
        self.FOVindex = index    
        
        # it updates the fov in the plot with the new index.
        self.ChangeFOV()
                
        
    def ChangeFOV(self):
        """
        it changes the fov or channel according to the choice of the user
        and it updates the plot shown and it initializes the new fov/channel
        at t=0 by default.
        """
        
        self.Tindex = 0
        
        # load the image and mask for the current plot
        self.m.currpicture = self.reader.LoadOneImage(self.Tindex,self.FOVindex)
        self.m.plotmask = self.reader.LoadMask(self.Tindex,self.FOVindex)
        
        # sets the image and the mask to 0 for the previous plot
        self.m.prevpicture = np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
        self.m.prevplotmask = np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
        
        # load the image and the mask for the next plot, check if it exists
        if self.Tindex+1 < self.reader.sizet:
            self.m.nextpicture = self.reader.LoadOneImage(self.Tindex+1, self.FOVindex)
            self.m.nextplotmask = self.reader.LoadMask(self.Tindex+1, self.FOVindex)
            
            # enables the next frame button in case it was disabled when the 
            # fov/channel was changed
            self.button_nextframe.setEnabled(True)
        else:
            self.m.nextpicture = np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
            self.m.nextplotmask =  np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
            
            # disables the next frame button if the mask or the picture
            # does not exist.
            self.button_nextframe.setEnabled(False)
            
        # once the images and masks are loaded into the variables, they are 
        # displaye in the gui.
        self.m.UpdatePlots()
        
        # disables the previous frame button in case it was active before 
        # changing fov/channel.
        self.button_previousframe.setEnabled(False)
        
        # updates the title of the plots to display the right time indices
        # aboves the plots.
        self.UpdateTitleSubplots()
            
        # if the button to hide the mask was checked before changing fov/channel,
        # it hides the mask again.
        if self.button_hidemask.isChecked():
            self.m.HideMask()
        
        # the button to set the time index is also set to 0/default again.
        self.button_timeindex.setText('')
        # enables the neural network buttons if there is already an 
        # existing prediction for the current image.
        self.EnableCNNButtons()
        
        
    def ReloadThreeMasks(self):
        """
        A function which replots all the masks at the current time and fov 
        indices. Needed after the batch prediction is completed to display
        the result of the NN.
        """
        log.debug("reload three mask for frame {}".format(self.Tindex))
        
        if self.Tindex >= 0 and self.Tindex <= self.reader.sizet-1:
            if self.Tindex == 0:
                self.button_nextframe.setEnabled(True)
                
                if self.Tindex < self.reader.sizet-1:
                    self.m.nextplotmask = self.reader.LoadMask(self.Tindex+1, self.FOVindex)
                else:
                    np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
                
                self.m.plotmask = self.reader.LoadMask(self.Tindex, self.FOVindex)
                self.m.prevplotmask = np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
                self.m.UpdatePlots()
                self.button_previousframe.setEnabled(False)
                
                
            elif self.Tindex == self.reader.sizet-1:
                self.button_previousframe.setEnabled(True)
                self.m.prevplotmask = self.reader.LoadMask(self.Tindex-1, self.FOVindex)
                self.m.plotmask = self.reader.LoadMask(self.Tindex, self.FOVindex)
                self.m.nextplotmask =  np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
                self.m.UpdatePlots()
                self.button_nextframe.setEnabled(False)
                
            else:
                self.button_nextframe.setEnabled(True)
                self.button_previousframe.setEnabled(True)
                self.m.prevplotmask = self.reader.LoadMask(self.Tindex-1, self.FOVindex)
                self.m.plotmask = self.reader.LoadMask(self.Tindex, self.FOVindex)              
                self.m.nextplotmask = self.reader.LoadMask(self.Tindex+1, self.FOVindex)
                self.m.UpdatePlots()
            
            self.UpdateTitleSubplots()
                        
            if self.button_hidemask.isChecked():
                self.m.HideMask()
            self.EnableCNNButtons()
        
        else:
            return
        
    
    def ChangeTimeFrame(self):
        """This funcion is called whenever the user gives a new time index, 
        to jump to the new given index, once "enter" button is pressed.
        """
        
        # it reads out the text in the button and converts it to an int.
        newtimeindex = int(self.button_timeindex.text())
        if newtimeindex >= 0 and newtimeindex <= self.reader.sizet-1:
            self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)
            
            self.Tindex = newtimeindex
            
            if self.Tindex == 0:
                self.button_nextframe.setEnabled(True)
                self.m.nextpicture = self.reader.LoadOneImage(self.Tindex+1,self.FOVindex)
                self.m.nextplotmask = self.reader.LoadMask(self.Tindex+1, self.FOVindex)
                
                self.m.currpicture = self.reader.LoadOneImage(self.Tindex, self.FOVindex)
                self.m.plotmask = self.reader.LoadMask(self.Tindex, self.FOVindex)
                
                self.m.prevpicture = np.zeros([self.reader.sizey, self.reader.sizex], 
                                              dtype = np.uint16)
                self.m.prevplotmask = np.zeros([self.reader.sizey, self.reader.sizex], 
                                               dtype = np.uint16)
    
                self.m.UpdatePlots()
                self.button_previousframe.setEnabled(False)
                
            elif self.Tindex == self.reader.sizet-1:
                self.button_previousframe.setEnabled(True)
                self.m.prevpicture = self.reader.LoadOneImage(self.Tindex-1, self.FOVindex)
                self.m.prevplotmask = self.reader.LoadMask(self.Tindex-1, self.FOVindex)
                   
                self.m.currpicture = self.reader.LoadOneImage(self.Tindex, self.FOVindex)
                self.m.plotmask = self.reader.LoadMask(self.Tindex, self.FOVindex)
                  
                self.m.nextpicture =  np.zeros([self.reader.sizey, self.reader.sizex], 
                                               dtype = np.uint16)
                self.m.nextplotmask =  np.zeros([self.reader.sizey, self.reader.sizex], 
                                                dtype = np.uint16)
                
                self.m.UpdatePlots()
                self.button_nextframe.setEnabled(False)
                
            else:
                self.button_nextframe.setEnabled(True)
                self.button_previousframe.setEnabled(True)
                self.m.prevpicture = self.reader.LoadOneImage(self.Tindex-1, self.FOVindex)
                self.m.prevplotmask = self.reader.LoadMask(self.Tindex-1, self.FOVindex)
                   
                self.m.currpicture = self.reader.LoadOneImage(self.Tindex, self.FOVindex)
                self.m.plotmask = self.reader.LoadMask(self.Tindex, self.FOVindex)              
                  
                self.m.nextpicture = self.reader.LoadOneImage(self.Tindex+1,self.FOVindex)
                self.m.nextplotmask = self.reader.LoadMask(self.Tindex+1, self.FOVindex)
                self.m.UpdatePlots()
            
            self.UpdateTitleSubplots()
            self.button_timeindex.clearFocus()
            self.button_timeindex.setText(str(self.Tindex)+'/'+str(self.reader.sizet-1))
            
            if self.button_hidemask.isChecked():
                self.m.HideMask()
            self.EnableCNNButtons()
        
        else:
            self.button_timeindex.clearFocus()
            return
        
    def BatchCellCorresp(self):
        def reset():
            self.ClearStatusBar()
            self.Enable(self.button_cellcorrespondence)
            self.button_cellcorrespondence.setChecked(False)
        
        self.Disable(self.button_cellcorrespondence)
        self.WriteStatusBar('retracking... ')

        if(self.Tindex==self.reader.sizet-1):
            msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Error',"This is the last frame. Nothing to retrack after this", parent=self)
            msg_box.exec()
            reset()
            return 


        # creates a dialog window from the BatchRetrack.py file
        dlg = br.CustomDialog(self)
        
        # this if tests if the user pressed 'ok' in the dialog window
        result = dlg.exec()
        if result == QDialog.DialogCode.Accepted:
            # it tests if the user has entered some values
            if not (dlg.entry1.text()!= ''):
                msg_box = QMessageBox(QMessageBox.Icon.Critical, "Error", "No Time Specified", parent=self)
                msg_box.exec()
                reset()
                return 
            
            # reads out the entry given by the user and converts the index
            # to integers
            time_value1 = int(dlg.entry1.text())
    
            # it tests if the value is bigger than current frame
            if (time_value1 <  self.Tindex or time_value1 > self.reader.sizet-1):
                msg_box = QMessageBox(QMessageBox.Icon.Critical, "Error", 'Invalid Time Constraints', parent=self)
                msg_box.exec()
                reset()
                return
            # if everything was okay, start a progress bar to show progress:
            progress = ProgressBar(self)
            progress_value = 0
            ratio = 100/(time_value1 - self.Tindex)
            for t in range(self.Tindex+1, time_value1+1):
                log.debug('start correspondance for frame {}'.format(t))                    
                #calls the cell correspondance for current time, t, and t+1
                temp_mask = self.reader.CellCorrespondence(t, self.FOVindex)
                progress_value += ratio
                progress.update_progress(progress_value)
                # update the progress status label
                progress.status.setText("Processing frame {}... ".format(t+1))
                self.reader.SaveMask(t, self.FOVindex, temp_mask)
                QApplication.processEvents()  # Process events to allow GUI to update               
                log.debug('finish correspondance and savemask for frame {}'.format(t))                    
            # close the progress bar dialog
            progress.close()
            self.ReloadThreeMasks()

            log.info('reload three frames')
        reset()

    def CellCorrespActivation(self):
        log.debug('An outdated function is called')
        self.Disable(self.button_cellcorrespondence)
        self.WriteStatusBar('Doing the cell correspondence')
        if self.Tindex > 0:
            self.m.plotmask = self.reader.CellCorrespondence(self.Tindex, self.FOVindex)
            self.m.updatedata()
        else:
            pass
            #self.m.plotmask = self.reader.LoadSeg(self.Tindex, self.FOVindex)
            #self.m.updatedata()

        self.Enable(self.button_cellcorrespondence)
        self.button_cellcorrespondence.setChecked(False)
        self.ClearStatusBar()
        self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)
    
    def ButtonSaveSegMask(self):
        """saves the segmented mask
        """
        self.reader.SaveSegMask(self.Tindex, self.FOVindex, self.m.plotmask)

        
    def ChangePreviousFrame(self):
         """This function is called when the previous frame buttons is pressed 
         and it tests if the buttons is enabled and if so it calls the
         BackwardTime() function. It should avoid the let the user do multiple 
         clicks and that the function is then called afterwards several times,
         once the frames and masks of the current time index have been loaded.
         """
         if self.button_previousframe.isEnabled():
            self.button_previousframe.setEnabled(False)
            self.BackwardTime()
            if self.Tindex >0:
                self.button_previousframe.setEnabled(True)
         else:
             return
             
             
    def ChangeNextFrame(self):
        """This function is called when the next frame buttons is pressed 
        and it tests if the buttons is enabled and if so it calls the
        ForwardTime() function. It should avoid the let the user do multiple 
        clicks and that the function is then called afterwards several times,
        once the frames and masks of the current time index have been loaded.
        """
        if self.button_nextframe.isEnabled():
            self.button_nextframe.setEnabled(False)
            self.ForwardTime()
            if self.Tindex + 1 < self.reader.sizet:
                self.button_nextframe.setEnabled(True)
                
        else:
            return

        
    def ForwardTime(self):
        """This function switches the frame in forward time index. And it tests
        several conditions if t == lastTimeIndex-1, because then the next frame
        button has to be disabled. It also tests if the show value of cells
        button and hidemask are active in order to hide/show the mask or to 
        show the cell values.
        """
        # the t frame is defined as the currently shown frame on the display.
        # If the button "Next time frame" is pressed, this function is called
        self.WriteStatusBar('Loading the next frame...')
        self.Disable(self.button_nextframe)

        if self.Tindex + 1 < self.reader.sizet - 1 :
            self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)
            
            self.m.prevpicture = self.m.currpicture.copy()
            self.m.prevplotmask = self.m.plotmask.copy()
            
            self.m.currpicture = self.m.nextpicture.copy()
            self.m.plotmask = self.m.nextplotmask.copy()
            
            self.m.nextpicture = self.reader.LoadOneImage(self.Tindex+2, self.FOVindex)
            self.m.nextplotmask = self.reader.LoadMask(self.Tindex+2, self.FOVindex)

            if self.Tindex + 1 == 1:
                self.button_previousframe.setEnabled(True)
                
        else:
            self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)
        
            self.m.prevpicture = self.m.currpicture.copy()
            self.m.prevplotmask = self.m.plotmask.copy()
            self.m.currpicture = self.m.nextpicture.copy()
            self.m.plotmask = self.m.nextplotmask.copy()
            self.m.nextpicture = np.zeros([self.reader.sizey, self.reader.sizex], 
                                          dtype = np.uint16)
            self.m.nextplotmask = np.zeros([self.reader.sizey,self.reader.sizex], 
                                           dtype = np.uint16)

            self.button_nextframe.setEnabled(False)

        self.Tindex = self.Tindex+1
        self.m.UpdatePlots()
        self.UpdateTitleSubplots()
        
        if self.button_hidemask.isChecked():
            self.m.HideMask()

        self.Enable(self.button_nextframe)
        self.ClearStatusBar()
        self.button_timeindex.setText(str(self.Tindex)+'/'+str(self.reader.sizet-1))

    
    def BackwardTime(self):
        """This function switches the frame in backward time index. And it 
        several conditions if t == 1, because then the button previous frame has to
        be disabled. It also tests if the show value of cells button and 
        hidemask are active in order to hide/show the mask or to show the cell
        values.
        """
        # the t frame is defined as the currently shown frame on the display.
        # If the button "Previous time frame" is pressed, this function is called
        self.WriteStatusBar('Loading the previous frame...')
        self.Disable(self.button_previousframe)
        
        self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)

        self.m.nextpicture = self.m.currpicture.copy()
        self.m.nextplotmask = self.m.plotmask.copy()
        self.m.currpicture = self.m.prevpicture.copy()
        self.m.plotmask = self.m.prevplotmask.copy()
            
        if self.Tindex == 1:
            self.m.prevpicture = np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
            self.m.prevplotmask = np.zeros([self.reader.sizey, self.reader.sizex], dtype = np.uint16)
            self.button_previousframe.setEnabled(False)
            
        else:
            self.m.prevpicture = self.reader.LoadOneImage(self.Tindex-2, self.FOVindex)
            self.m.prevplotmask = self.reader.LoadMask(self.Tindex-2, self.FOVindex)

        
        if self.Tindex-1 == self.reader.sizet-2:
            self.button_nextframe.setEnabled(True)            
        
        if self.button_hidemask.isChecked():
            self.m.HideMask()
        
        self.Tindex -= 1
        self.m.UpdatePlots()
        self.UpdateTitleSubplots()
            
        self.Enable(self.button_previousframe)
        
        if self.Tindex > 0:
            self.button_previousframe.setEnabled(True)          
            
        self.ClearStatusBar()
        self.button_timeindex.setText(str(self.Tindex)+'/' + str(self.reader.sizet-1))


# -----------------------------------------------------------------------------
# MANUAL MASK CORRECTIONS
            
    def ChangeOneValue(self):
        """This function is called when the button Change cell value is
        clicked. It displays the instructions on the status bar.
        And if the user clicks in the graph where the current mask is displayed
        it connects the event of the click (meaning that user has clicked on
        one cell) to the function self.DialogBoxChangeOneValue. 
        This function will then replaces the cell selected by the user with
        the click with a new value entered by the user.
        """
        
        # displaying the instructions on the statusbar
        self.WriteStatusBar((
            'Left-click to select cell, right-click to abort.'))

        # disables all the buttons
        self.Disable(self.button_changecellvalue)
        
        # connects the event "press mouse button" in the matplotlib plot 
        # (picture) to the function self.DialogBoxChangeOneValue
        self.id = self.m.mpl_connect('button_press_event', self.DialogBoxChangeOneValue)
        
        
    def DialogBoxChangeOneValue(self, event):
        """This function is called when the user after the user has selected
        the button Change cell value and clicked in the picture to select
        the desired cell to change.
        
        It first deconnects the mouse click event in matplotlib with this 
        function to not generate any other dialog window.
        
        It then tests if the click is inside the matplotlib plot (if it is
        outside it equals to None) and if it is the current and editable plot
        (the one in the middle of the gui, self.m.ax)
        
        If is true, then it sets the coordinates to int. and creates a dialog
        window where the user is asked to type a value to set it to the cell.
        
        If the user presses ok, it tests if the entry is valid (>0 and not 
        empty) and looks for the old cell value and replaces it. And then
        it updates the plot such that the result of the change can be seen.
        """
        # the function is disconnected from the matplotlib event.
        self.m.mpl_disconnect(self.id)
        
        # test if the button is a left click and if the coordinates
        # chosen by the user click is inside of the current matplotlib plot
        # which is given by self.m.ax
        if (event.button == 1 
            and (event.xdata != None and event.ydata != None) 
            and self.m.ax == event.inaxes):
            newx = int(event.xdata)
            newy = int(event.ydata)
            
            # creates a dialog window
            dlg = cocv.CustomDialog(self)
            
            #if the user presses 'ok' in the dialog window it executes the code
            #else it does nothing
            result = dlg.exec()
            if result == QDialog.DialogCode.Accepted:
                #it tests that the user has entered some value, that it is not
                #empty and that it is equal or bigger to 0.
                if dlg.entry1.text() != '' and int(dlg.entry1.text()) >= 0:
                    #reads the new value to set and converts it from str to int
                    value = int(dlg.entry1.text())
                    
                    # gives the coordinates where it is equal to the value
                    # selected by the user. And it replaces it with the new
                    # value.
                    self.m.plotmask[self.m.plotmask == self.m.plotmask[newy,newx]] = value
                    
                    # updates the plot to see the modification.
                    self.m.updatedata()
                    
        self.Enable(self.button_changecellvalue)
        self.button_changecellvalue.setChecked(False)
        self.m.ShowCellNumbers(type='current')
        self.SaveMask()
        self.ClearStatusBar()
        
        
    def DialogBoxECV(self, s):
        """This functions creates from the ExchangeCellValues.py file a 
        window which takes two integer entries and then swaps the cells having
        the given integer values.
        """
        # creates a dialog window from the ExchangeCellValues.py file
        dlg = ecv.CustomDialog(self)
        
        # if the user presses 'ok', it executes the code
        result = dlg.exec()
        if result == QDialog.DialogCode.Accepted:
            # it tests if both value to be swapped are not empty.
            if dlg.entry1.text()!= '' and dlg.entry2.text() != '':
                
                # reads out the values and converts it into integers.
                value1 = int(dlg.entry1.text())
                value2 = int(dlg.entry2.text())
                
                # calls the function which does the swap
                try:
                    self.m.ExchangeCellValue(value1,value2)
                except ValueError as e:
                    msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Error', str(e), parent=self)
                    msg_box.exec()
                self.m.ShowCellNumbers(type='current')
                self.SaveMask()
        else:
            return


    def MouseDraw(self):
        """
        This function is called whenever the brush or the eraser button is
        pressed. On the first press event it calls the self.m.OneClick, which
        tests whether it is a right click or a left click. If it a right click
        it assigns the value of the pixel which has been right clicked
        to self.cellval, meaning that the next drawn pixels will be set to this
        value.
        If it is left clicked, then it draws a 3x3 square with the current
        value of self.cellval.
        If after left clicking you drag the mouse, then you start drawing 
        using the mouse and it stops once you release the left click.
        
        Same for the eraser button, it sets directly the value of self.cellval
        to 0.
        """
        do_draw = self.button_drawmouse.isChecked()
        do_erase = self.button_eraser.isChecked()
        
        if do_draw or do_erase:
            self.m.tempmask = self.m.plotmask.copy()
            
            #save status of the showval check box, then turn off (slows brush/eraser down too much)
            self.showval_was_checked = self.button_showval.isChecked()
            self.button_showval.setChecked(False)
            
            if do_draw:
                self.WriteStatusBar(('Draw using the brush, right click to select '
                                     'the cell to draw.'))
                self.Disable(self.button_drawmouse)
                
            elif do_erase:
                self.WriteStatusBar('Erasing by setting the values to 0.')
                self.Disable(self.button_eraser)
                self.m.cellval = 0
                
            radius = self.spinbox_brushsize.value()
            self.id2 = self.m.mpl_connect('button_press_event', 
                                          lambda e: self.m.OneClick(e, radius))
            self.id = self.m.mpl_connect('motion_notify_event', 
                                          lambda e: self.m.PaintBrush(e, radius))
            self.id3 = self.m.mpl_connect('button_release_event', self.m.ReleaseClick)
                
                        
        else:
            self.m.mpl_disconnect(self.id3)
            self.m.mpl_disconnect(self.id2)
            self.m.mpl_disconnect(self.id)
            QApplication.restoreOverrideCursor()
            self.Enable(self.button_drawmouse)
            self.Enable(self.button_eraser)
            self.button_showval.setChecked(self.showval_was_checked)
            self.SaveMask()
            self.ClearStatusBar()
            
            
    
    def SelectSplitCell(self):
        """Function called upon clicking Split Cell button"""
      
        if self.button_split.isChecked():
            self.cell_to_split=-1
            self.Disable(self.button_split)
            self.WriteStatusBar('First, select the cell to be split with a left click. '
                                'Then, left-click the corners of a polygon to be split off from the '
                                'selected cell. Right click to confirm.')
            
            def sel_cell(e):
                if e.button != 1:
                    self.cell_to_split=-1
                    self.Enable(self.button_split)
                else:
                    x = int(e.xdata)
                    y = int(e.ydata)
                    self.cell_to_split = self.m.plotmask[y,x]
                    self.m.mpl_disconnect(self.id)
                    self.ClickSplitCell()
    
            self.id = self.m.mpl_connect('button_press_event', sel_cell)
            
        else:
            self.DoSplitCell()


    def ClickSplitCell(self):
        self.WriteStatusBar('Click to define the corners of a polygon which'
                            'will be split off the selected cell. Right click '
                            'or reclick the button to confirm.')
        if self.cell_to_split == -1:
            return
        
        self.m.tempmask = self.m.plotmask.copy()
        self.m.storemouseclicks = [] 
        self.id = self.m.mpl_connect('button_press_event', 
                                     lambda e: self.m.multiple_click(e, self.DoSplitCell))

        
    def DoSplitCell(self):
        self.m.mpl_disconnect(self.id)
        if len(self.m.storemouseclicks) > 2:
            nx, ny = self.m.plotmask.shape
            img = Image.new('L', (ny, nx), 0)
            ImageDraw.Draw(img).polygon(self.m.storemouseclicks, outline=1, fill=1)
            polygon = np.array(img).astype(bool)
            selected_mask = self.m.plotmask==self.cell_to_split
            replace_cell = self.m.plotmask.max() + 1
            self.m.plotmask[selected_mask & polygon] = replace_cell
        self.Enable(self.button_split)
        self.m.UpdatePlots(type='current')
        self.ClearStatusBar()
        self.SaveMask()
        self.m.storemouseclicks = []
        self.button_split.setChecked(False)
        
        
    def UpdateTitleSubplots(self):
        """This function updates the title of the plots according to the 
        current time index. So it called whenever a frame or a fov is changed.
        """
        if self.Tindex == 0:
            self.m.titlecurr.set_text('Time index {}'.format(self.Tindex))
            self.m.titleprev.set_text('No frame {}'.format(''))
            self.m.titlenext.set_text('Next time index {}'.format(self.Tindex+1))
            self.m.draw()
            
        elif self.Tindex == self.reader.sizet-1:
            self.m.titlecurr.set_text('Time index {}'.format(self.Tindex))
            self.m.titleprev.set_text('Previous time index {}'.format(self.Tindex-1))
            self.m.titlenext.set_text('No frame {}'.format(''))            
            self.m.draw()
            
        else:
            self.m.titlecurr.set_text('Time index {}'.format(self.Tindex))
            self.m.titleprev.set_text('Previous time index {}'.format(self.Tindex-1))
            self.m.titlenext.set_text('Next time index {}'.format(self.Tindex+1))
            self.m.draw()
        
        
    def ClickNewCell(self):
        """ 
        this method is called when the button New Cell is clicked. If the button
        state corresponds to True (if is activated) then it connects the mouse 
        clicks on the pyqt window to the canvas (so to the matplolib figure).
        The connection has an "id" which is given by the integer self.id
        After the connections is made, it calls the Disable function with argument 0
        which turns off the other button(s).
        
        If the button is clicked but it is deactivated then it disconnects the 
        connection between the canvas and the window (the user can not interact
        with the plot anymore). 
        Storemouseclicks is a list corresponding to the coordinates of all mouse
        clicks between the activation and the deactivation of the button.
        So if it is empty, it does not draw anything because no clicks
        were registered. 
        But if it has some coordinates, it will draw a polygon where the vertices
        are the coordinates of all the mouseclicks.
        Once the figure has been updated with a new polygon, the other button(s)
        are again enabled. 
        
        """
        if self.button_newcell.isChecked():
            self.WriteStatusBar(('Draw a new cell. Use the left click to '
                                 'produce a polygon with a new cell value. '
                                 'Click the button again to confirm.'))
            self.m.tempmask = self.m.plotmask.copy()
            self.id = self.m.mpl_connect('button_press_event', self.m.MouseClick)
            self.Disable(self.button_newcell)
            
        else:
            self.m.mpl_disconnect(self.id)
            if  self.m.storemouseclicks and self.TestSelectedPoints():
                self.m.DrawRegion(True)
            else:
                self.m.updatedata()
            self.Enable(self.button_newcell)
            self.m.ShowCellNumbers(type='current')
            self.ClearStatusBar()
            self.SaveMask()
            
            
    def TestSelectedPoints(self):
        """This function is just used to catch an exception, when the new cell
        or the add region function is called. If all the dots drawn by the user
        are located on one line (horizontal or vertical) the DrawRegion 
        function calls a method to create a polygon and 
        it can not make a polygon out of straight line so it gives an error.
        In order to prevent this error, this function avoids to attempt to draw
        by returning False if the square are all on one line.
        """
                
        allx = list(np.array(self.m.storemouseclicks)[:,0])
        ally = list(np.array(self.m.storemouseclicks)[:,1])
        
        resultx = all(elem == allx[0] for elem in allx)
        resulty = all(elem == ally[0] for elem in ally)
        
        if resultx or resulty:
            return False
        else:
            return True
                    
        
    def clickmethod(self):
        """ 
        this method is called when the button Add region is clicked. If the button
        state corresponds to True (if is activated) then it connects the mouse 
        clicks on the pyqt window to the canvas (so to the matplolib figure).
        The connection has an "id" which is given by the integer self.id
        After the connections is made, it calls the Disable function with argument 1
        which turns off the other button(s).
        
        If the button is clicked and it is deactivated then it disconnects the 
        connection between the canvas and the window (the user can not interact
        with the plot anymore). 
        Storemouseclicks is a list corresponding to the coordinates of all mouse
        clicks between the activation and the deactivation of the button.
        So if it is empty, it does not draw anything because no clicks
        were registered. 
        But if it has some coordinates, it will draw a polygon where the vertices
        are the coordinates of all the mouseclicks.
        Once the figure has been updated with a new polygon, the other button(s)
        are again enabled. 
        
        """
        if self.button_add_region.isChecked():
            self.WriteStatusBar(('Select the cell with the first click, '
                                 'then draw polygon with subsequent '
                                 'clicks. Reclick the button to confirm.'))
            self.m.tempmask = self.m.plotmask.copy()
            self.id = self.m.mpl_connect('button_press_event', self.m.MouseClick)           
            self.Disable(self.button_add_region) 
            
        else:
            self.m.mpl_disconnect(self.id)
            
            # test if the list is not empty and if the dots are not all in the same line
            if self.m.storemouseclicks and self.TestSelectedPoints():
                self.m.DrawRegion(False)

            else:
                self.m.updatedata()
                
            self.Enable(self.button_add_region)
            self.m.ShowCellNumbers(type='current')
            self.ClearStatusBar()
            self.SaveMask()
            
            
    def MergeWithNeighbors(self):
        """This function is called when the button Merge With Neighbors is
        clicked. It displays the instructions on the status bar.
        And if the user clicks in the graph where the current mask is displayed
        it connects the event of the click (meaning that user has clicked on
        one cell) to the function self.PerformMergeWithNeighbors. 
        This function will then merge the cell selected by the user with
        the click with all (touching) neighbors.
        """
        
        # displaying the instructions on the statusbar
        self.WriteStatusBar((
            'Left-click to select cell, right-click to abort.'))

        # disables all the buttons
        self.Disable(self.button_mergewithneighbors)
        
        # connects the event "press mouse button" in the matplotlib plot 
        # (picture) to the function self.PerformMergeWithNeighbors
        self.id = self.m.mpl_connect('button_press_event', self.PerformMergeWithNeighbors)
        
        
    def PerformMergeWithNeighbors(self, event):
        """This function is called after the user has selected
        the button Merge With Neighbors and clicked in the picture to select
        the desired cell to start the merger.
        
        It first deconnects the mouse click event in matplotlib with this 
        function to not generate any other dialog window.
        
        It then tests if the click is inside the matplotlib plot (if it is
        outside it equals to None) and if it is the current and editable plot
        (the one in the middle of the gui, self.m.ax)
        
        If is true, then it sets the coordinates to int. and creates a dialog
        window where the user is asked to type a value to set it to the cell.
        
        If the user presses ok, it tests if the entry is valid (>0 and not 
        empty) and looks for the old cell value and replaces it. And then
        it updates the plot such that the result of the change can be seen.
        """
        # the function is disconnected from the matplotlib event.
        self.m.mpl_disconnect(self.id)
        
        # test if the button is a left click and if the coordinates
        # chosen by the user click is inside of the current matplotlib plot
        # which is given by self.m.ax
        if (event.button == 1 
            and (event.xdata != None and event.ydata != None) 
            and self.m.ax == event.inaxes):
            newx = int(event.xdata)
            newy = int(event.ydata)
            # get the ID of the selected cell
            selectedcell_ID = self.m.plotmask[newy,newx]
            # binarize the existing mask and number ("label") each component
            labeled_binarized_mask = skimage.measure.label(self.m.plotmask > 0)
            # select the component pixels that corresponds to the selected point
            flood_indices=(labeled_binarized_mask==labeled_binarized_mask[newy,newx])
            # set the mask pixels that correspond to the component to the cell ID
            self.m.plotmask[flood_indices] = selectedcell_ID
                    
            # updates the plot to see the modification.
            self.m.updatedata()
                    
        self.Enable(self.button_mergewithneighbors)
        self.button_mergewithneighbors.setChecked(False)
        self.m.ShowCellNumbers(type='current')
        self.SaveMask()
        self.ClearStatusBar()
        
        
# -----------------------------------------------------------------------------
# BUTTON ENABLE / DISABLE
    
    def Enable(self, button):
        """
        this functions turns on buttons all the buttons, depending on the time
        index. (next and previous buttons should not be turned on if t = 0 
        or t = lasttimeindex)
        """
        for k in range(0, len(self.buttonlist)):
            if button != self.buttonlist[k]:
                self.buttonlist[k].setEnabled(True)
                
        if self.Tindex == 0:
            self.button_previousframe.setEnabled(False)
            
        if self.Tindex == self.reader.sizet-1:
            self.button_nextframe.setEnabled(False)
            
        self.EnableCNNButtons()
         
     
    def Disable(self, button):
        """
        this functions turns off all the buttons except the one given in 
        argument.
        """

        for k in range(0,len(self.buttonlist)):
            if button != self.buttonlist[k]:
                self.buttonlist[k].setEnabled(False)


    def EnableCNNButtons(self):
        if self.reader.TestTimeExist(self.Tindex, self.FOVindex):
            self.button_cellcorrespondence.setEnabled(True)
            self.button_extractfluorescence.setEnabled(True)
        else:
            self.button_cellcorrespondence.setEnabled(False)
            self.button_extractfluorescence.setEnabled(False)
    
    
    def EnableCorrectionsButtons(self):
        self.button_newcell.setEnabled(True)
        self.button_add_region.setEnabled(True)
        self.button_drawmouse.setEnabled(True)
        self.button_eraser.setEnabled(True)
        self.button_exval.setEnabled(True)
        self.button_changecellvalue.setEnabled(True)
        self.button_showval.setEnabled(True)
        self.button_split.setEnabled(True)
        
        
    def DisableCorrectionsButtons(self):
        self.button_newcell.setEnabled(False)
        self.button_add_region.setEnabled(False)
        self.button_drawmouse.setEnabled(False)
        self.button_eraser.setEnabled(False)
        self.button_exval.setEnabled(False)
        self.button_changecellvalue.setEnabled(False)
        self.button_showval.setEnabled(False)
        self.button_split.setEnabled(False)
        
    
    def SaveMask(self):
        """
        When this function is called, it saves the current mask
        (self.m.plotmask)
        """
        self.reader.SaveMask(self.Tindex, self.FOVindex, self.m.plotmask)
        
        
    def WriteStatusBar(self, text):
        """Writes text to status bar"""
        self.statusBarText.setText(text)
        
        
    def ClearStatusBar(self):
        """Removes text from status bar"""
        self.statusBarText.setText('')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
GUI to select and deselect cells at extraction time.
"""

import sys
import os
import numpy as np
from PyQt6.QtWidgets import (QApplication, QPushButton, QLabel,
                             QHBoxLayout, QVBoxLayout, QListWidget,
                             QFileDialog, QMessageBox, QDialog)
from PyQt6.QtCore import Qt


from PIL import Image, ImageDraw

#Import from matplotlib to use it to display the pictures and masks.
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

from matplotlib import cm
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.figure import Figure

from ..disk.image_loader import load_image

class Extract(QDialog):
    
    
    def __init__(self, image, mask, channel_names=[]):
        parent = None
        super(Extract, self).__init__(parent)
        self.setWindowFlags(Qt.WindowType.WindowStaysOnTopHint)
#        image, mask = _test_data()
        self.pc = PlotCanvas(image, mask)
        self.file_list = channel_names
        self.init_UI()
        self.exit_code = 0 # 0: Cancel, 1: Fluorescence, 2: Mask

    def init_UI(self):
        # Extract Buttons
        title_extr = QLabel('Extract:')
        self.extr_mask = _create_button("Extract mask", self.do_extr_mask)
        self.extr_fluo = _create_button("Extract values", self.do_extr_fluo)
        self.done = _create_button("Cancel", self.do_cancel)
        self.done.setDefault(True)  
        
        extr_box = QVBoxLayout()
        extr_box.addWidget(title_extr)
        extr_box.addWidget(self.extr_mask)
        extr_box.addWidget(self.extr_fluo)
        extr_box.addWidget(self.done)
        extr_box.setAlignment(Qt.AlignmentFlag.AlignTop)
        
        # Select Button
        title_select = QLabel('Select cells:')
        self.sel_mult = _create_button("Select multiple cells", self.do_sel_mult,
                                       "Left-click to define the corners of a polygon around cells, "
                                       "right-click to confirm.")
        self.sel_sngl = _create_button("Select single cell", self.do_sel_sngl,
                                       "Left-click to select cell, right-click "
                                       "to abort.")
        self.desel_mult = _create_button("Deselect multiple cells", self.do_desel_mult,
                                       "Left-click to define the corners of a polygon around cells, "
                                       "right-click to confirm.")
        self.desel_sngl = _create_button("Deselect single cell", self.do_desel_sngl,
                                       "Left-click to deselect cell, right-click "
                                       "to abort.")
        
        sel_box = QVBoxLayout()
        sel_box.addWidget(title_select)
        sel_box.addWidget(self.sel_mult)
        sel_box.addWidget(self.sel_sngl)
        sel_box.addWidget(self.desel_mult)
        sel_box.addWidget(self.desel_sngl)
        sel_box.setAlignment(Qt.AlignmentFlag.AlignTop)
        
        # Additional Fluorescence File
        file_title = QLabel("Manage files and channels for extraction:")
        self.list_channels = QListWidget()
        self.add_file = _create_button('Add', self.do_add_file)
        self.remove_file = _create_button('Remove', self.do_remove_file)
        
        add_remove = QHBoxLayout()
        add_remove.addWidget(self.add_file)
        add_remove.addWidget(self.remove_file)
        
        manage_box = QVBoxLayout()
        manage_box.addWidget(file_title)
        manage_box.addWidget(self.list_channels)
        manage_box.addLayout(add_remove)
        manage_box.setAlignment(Qt.AlignmentFlag.AlignTop)
        
        # Button list
        self.buttons = [self.extr_mask,
                        self.extr_fluo,
                        self.done,
                        self.sel_mult,
                        self.sel_sngl,
                        self.desel_mult,
                        self.desel_sngl,
                        self.add_file,
                        self.remove_file,
                        self.list_channels]
        
        # Buttons
        buttons = QHBoxLayout()
        buttons.addLayout(sel_box)
        buttons.addLayout(manage_box)
        buttons.addLayout(extr_box)
        
        # Plot Canvas
        full = QVBoxLayout()
        full.addWidget(self.pc)
        full.addStretch(.5)
        full.addLayout(buttons)
        
        # Add channel list
        self.do_show_list()

        self.setLayout(full)    
        self.setGeometry(300, 300, 600, 600)
        self.setWindowTitle('Extract geometries, fluorescence, masks')
        self.show()

    def do_extr_fluo(self):
        self.outfile, _ = QFileDialog.getSaveFileName(
            self,"Specify CSV file for exporting values",
            "","All files (*);;Text files (*.csv)")
        _, ext = os.path.splitext(self.outfile)
        if ext == '':
            self.outfile += '.csv'
        elif ext != '.csv':
            msg_box = QMessageBox(QMessageBox.Icon.Critical,'Error','Must specify .csv file', parent=self)
            msg_box.exec()
            return 
        
        self.exit_code = 1
        self.cells = self.pc.sellist
        self.desel_cells = set(np.unique(self.pc.mask)) - set(self.cells)
        self.close()

    def do_cancel(self):
        self.close()
    
    def do_extr_mask(self):
        self.outfile, _ = QFileDialog.getSaveFileName(
            self,"Specify TIFF file for exporting mask",
            "","All files (*);;Image file (*.tiff)")
        _, ext = os.path.splitext(self.outfile)
        if ext == '':
            self.outfile += '.tif'
        elif ext != '.tif' and ext!='.tiff' and ext!='.TIF' and ext!='.TIFF':
            msg_box = QMessageBox(QMessageBox.Icon.Critical,'Error','Must specify .tif file', parent=self)
            msg_box.exec()
            return 

        self.exit_code = 2
        self.cells = self.pc.sellist
        self.desel_cells = set(np.unique(self.pc.mask)) - set(self.cells)
        self.close()
                
    def do_sel_mult(self):
        """Callback for selecting multiple"""
        self.pc.storemouseclicks = []        
        def to_connect(e):
            self.pc.multiple_click(e, self.do_sel_mult_process)
        self.pc.connect_id = self.pc.mpl_connect(
                'button_press_event', 
                 to_connect)
        self.deactivate_all()
        
    def do_sel_mult_process(self):
        """Process selecting multiple"""
        cells = self.cells_in_polygon()
        self.pc.sellist = self.pc.sellist.union(cells)
        self.disconnect()
        
    def do_desel_mult(self):
        self.pc.storemouseclicks = []        
        """Callback for deselecting multiple"""
        def to_connect(e):
            self.pc.multiple_click(e, self.do_desel_mult_process)
        self.pc.connect_id = self.pc.mpl_connect(
                'button_press_event', 
                 to_connect)
        self.deactivate_all()
        
    def do_desel_mult_process(self):
        """Process deselect multiple"""
        cells = self.cells_in_polygon()
        self.pc.sellist = self.pc.sellist - cells
        self.disconnect()
        
    def cells_in_polygon(self):
        """Extracts cells inside of polygon specified by pc.storemouseclicks"""
        nx, ny = self.pc.mask.shape
        img = Image.new('L', (ny, nx), 0)
        ImageDraw.Draw(img).polygon(self.pc.storemouseclicks, outline=1, fill=1)
        polygon = np.array(img).astype(bool)
        return set(np.unique(self.pc.mask[polygon]))
    
    def do_sel_sngl(self):
        """Select single cell"""
        def to_connect(e):
            self.pc.single_click(e, self.do_sel_sngl_process)
        self.pc.connect_id = self.pc.mpl_connect(
                'button_press_event', 
                 to_connect)
        self.deactivate_all()
    
    def do_sel_sngl_process(self, x, y):
        """Process selected cell"""
        if x is not None:
            cell = self.pc.mask[y,x]
            self.pc.sellist.add(cell)
        self.disconnect()
    
    def do_desel_sngl(self):
        """Deselect single cell"""
        def to_connect(e):
            self.pc.single_click(e, self.do_desel_sngl_process)
        self.pc.connect_id = self.pc.mpl_connect(
                'button_press_event', 
                 to_connect)
        self.deactivate_all()
        
    def do_desel_sngl_process(self, x, y):
        """Process deselected cell"""
        if x is not None:
            cell = self.pc.mask[y,x]
            self.pc.sellist.remove(cell)
        self.disconnect()
            
    def disconnect(self):
        """Disconnects callback, updates plots, activates buttons"""
        self.pc.update_plots()
        self.pc.mpl_disconnect(self.pc.connect_id)
        self.activate_all()
        
    def deactivate_all(self):
        for b in self.buttons:
            b.setEnabled(False)
            
    def activate_all(self):
        for b in self.buttons:
            b.setEnabled(True)
            
    def do_add_file(self):
        dlg = QFileDialog()
        dlg.setProxyModel(None)
        if dlg.exec():
            full_files = dlg.selectedFiles()
            if self.test_file(full_files):
                self.file_list = self.file_list + full_files
                self.do_show_list()
    
    def do_remove_file(self):
        item_to_remove = self.list_channels.currentItem()
        remove_ix = self.list_channels.row(item_to_remove)
        self.file_list.pop(remove_ix)
        self.do_show_list()
    
    def do_show_list(self):
        self.list_channels.clear()
        for file in self.file_list:
            _, name = os.path.split(file)
            self.list_channels.addItem(name)
            
    def test_file(self, files):
        """Tests if input image has appropriate size and contains data"""
        for f in files:
            try: 
                im = load_image(f, 0)
            except ValueError:
                msg_box = QMessageBox(QMessageBox.Icon.Critical, "Error", "Could not load file", parent=self)
                msg_box.exec()
                return False
            if not im.shape == self.pc.mask.shape:
                msg_box = QMessageBox(QMessageBox.Icon.Critical, "Error", "Loaded image has wrong size", parent=self)
                msg_box.exec()
                return False
        return True
            
    

class PlotCanvas(FigureCanvas):
    
    
    def __init__(self, image, mask, parent=None, figsize=(5,4)):
        """this class defines the canvas. It initializes a figure, which is then
        used to plot our data using imshow.
        """
        fig = Figure()
        self.ax = fig.add_subplot(111)
        super(PlotCanvas, self).__init__(fig)
        self.setParent(parent)
        
        self.image = image
        self.mask = mask
        self.sellist = set(np.unique(mask)) # selected cells
        self.vismask = mask.copy() # mask with only selected cells
        
        self.ax_image, self.ax_mask = self.initialize_plots(image, mask, self.ax)        
        self.storemouseclicks = []
        self.connect_id = None

    def initialize_plots(self, picture, mask, ax):
        """Creates plots at initialization"""
        newcmp = _colormap(21)
        ax.axis("off")
        self.draw()
        return (ax.imshow(picture, interpolation= 'None', 
                          origin = 'upper', cmap = 'gray_r'), 
                ax.imshow((mask%10+1)*(mask != 0), origin = 'upper', 
                          interpolation = 'None', cmap = newcmp,
                          vmin=0, vmax=11))
            
    def single_click(self, event, call_after):
        """Function for single click, calls call_after afterwards with
        the clicked points (or None, None) if abort click."""
        if (event.button == 1
            and (event.xdata != None and event.ydata != None) 
            and self.ax == event.inaxes):
            x = int(event.xdata)
            y = int(event.ydata)
            call_after(x, y)
        else:
            call_after(None, None)
            
    def multiple_click(self, event, call_after):
        """Function to keep track of multiple left clicks, confirmed with 
        a right click. After right click, the function call_after is called"""        
        
        if (event.button == 1  # left click
            and (event.xdata != None and event.ydata != None) 
            and self.ax == event.inaxes):
            
            newx = int(event.xdata)
            newy = int(event.ydata)
            self.storemouseclicks.append((newx, newy))
            self.draw_click(newx, newy)
        
        elif (event.button == 3): # right click
            self.mpl_disconnect(self.connect_id)
            call_after()
        
    def draw_click(self, posx, posy):
        """
        it updates the plot once the user clicks on the plot and draws a 4x4 pixel dot
        at the coordinate of the click 
        """     
        self.vismask[posy:posy+2, posx:posx+2] = 9
        self.redraw_mask()

    def redraw_mask(self):
        """Redraw mask with current self.vismask"""
        self.ax_mask.set_data((self.vismask%10+1)*(self.vismask!=0))
        self.ax.draw_artist(self.ax_image)
        self.ax.draw_artist(self.ax_mask)
        self.update()
        self.flush_events()

    def recalculate_vismask(self):
        """Recalculates vismask with current list of cells to show"""
        tmp = self.mask.copy()
        all_cells = set(np.unique(tmp))
        to_remove = all_cells - self.sellist
        for cell in to_remove:
            tmp[tmp==cell] = 0
        self.vismask = tmp

    def update_plots(self):
        """Shows plot with currently selected cells"""
        self.recalculate_vismask()
        self.redraw_mask()

def _create_button(text, connect, tooltip=None):
    """Initializes button, connects with callback connect"""
    button = QPushButton(text)
    button.clicked.connect(connect)
    button.setMaximumWidth(150)
    
    if tooltip is not None:
        button.setToolTip(tooltip)
    
    return button

def _colormap(Ncolors=21):
    """Creates colormap for segmentation mask"""
    jet = cm.get_cmap('jet', Ncolors)
    
    cmaplist = [(0,0,0,0)]
    for i in range(jet.N):
        r,g,b,_ = jet(i)
        cmaplist.append((r,g,b,.2))
    cmap = LinearSegmentedColormap.from_list('Custom cmap', cmaplist, Ncolors)
    return cmap  
        
def _poly_to_mask(polygon, shape):
    """Converts polygon to mask"""
    img = Image.new('L', shape, 0)
    ImageDraw.Draw(img).polygon(polygon, outline=0, fill=1)
    return np.array(img).astype(int)

def _poly_to_line(polygon, shape):
    """Converts polygon to line"""
    img = Image.new('L', shape, 0)
    ImageDraw.Draw(img).polygon(polygon, outline=1, fill=0)
    return np.array(img).astype(int)

def _test_data():
    """Creates test data"""
    im = np.zeros((100,100))
    mask = np.zeros(im.shape)

    poly1 = [(20,20),(30,30),(20,40),(10,30)]
    poly2 = [(50,50),(60,60),(50,70),(40,60)]

    mask += _poly_to_mask(poly1, im.shape)
    mask += _poly_to_mask(poly2, im.shape)*2
    im += _poly_to_line(poly1, im.shape)
    im += _poly_to_line(poly2, im.shape)
    
    return im, mask
        
        
if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = Extract(*_test_data())
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 17:38:58 2019
"""

from PyQt6.QtWidgets import QApplication, QMainWindow, QMenu, QVBoxLayout, QSizePolicy, QMessageBox, QWidget, QPushButton, QComboBox, QDialog, QDialogButtonBox, QInputDialog, QLineEdit, QFormLayout
from PyQt6 import QtGui
from PyQt6.QtGui import QShortcut
from PyQt6.QtCore import pyqtSignal, QObject, Qt
#import PyQt package, allows for GUI interactions

class CustomDialog(QDialog):

    def __init__(self, *args, **kwargs):
        super(CustomDialog, self).__init__(*args, **kwargs)
        
        self.setWindowTitle("Change Value one cell")
        self.setGeometry(100,100, 500,200)
        
        self.entry1 = QLineEdit()
        self.entry1.setValidator(QtGui.QIntValidator())
        self.entry1.setMaxLength(4)
        self.entry1.setAlignment(Qt.AlignmentFlag.AlignRight)

        
#        self.entry2 = QLineEdit()
#        self.entry2.setValidator(QtGui.QIntValidator())
#        self.entry2.setMaxLength(4)
#        self.entry2.setAlignment(Qt.AlignmentFlag.AlignRight)
        
        flo = QFormLayout()
        flo.addRow('Enter Cell value (integer):', self.entry1)
#        flo.addRow('Enter Cell value 2 (integer):', self.entry2)        
        
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

#        self.layout = QVBoxLayout()
#        self.layout.addWidget(self.buttonBox
        flo.addWidget(self.buttonBox)
        self.setLayout(flo)
        
        # Set focus to entry1
        self.entry1.setFocus()
        


# -*- coding: utf-8 -*-
"""
Initializing all the buttons in this file.
"""
from PyQt6.QtWidgets import QPushButton, QComboBox, QLineEdit
from PyQt6 import QtGui
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QKeySequence

import numpy as np
import os
import sys

import importlib.resources as resources
# Access icon file
path_icons = resources.files('yeaz.icons')

def Init(parent):
    """
    configuration of all the buttons, some buttons just need to be toggled
    meaning that they need just to be clicked and the function associated 
    to the button is called and executed. Other buttons are checkable 
    meaning that until the user has not finished to use the function
    connected to the button, this button stays active (or Checked)
    """

    # hide / show cnn buttons
    parent.button_hide_show.setCheckable(True)    
    parent.button_hide_show.setMaximumWidth(150)
    parent.button_hide_show.clicked.connect(parent.ShowHideCNNbuttons)
    parent.button_hide_show.setStatusTip('Press to hide and show neural network buttons.')

    # ADD REGION
    parent.button_add_region.setCheckable(True)
    parent.button_add_region.setMaximumWidth(150)
    parent.button_add_region.clicked.connect(parent.clickmethod)
    parent.button_add_region.setShortcut("R")
    parent.button_add_region.setToolTip("Shortcut: R key")
    parent.button_add_region.setStatusTip('First left click identifies the ID value for the cell (or background) to be extended, then left-click to specify a polygon.')
    
    # NEW CELL
    parent.button_newcell.setCheckable(True)
    parent.button_newcell.setMaximumWidth(150)
    parent.button_newcell.setShortcut("N")
    parent.button_newcell.setToolTip("Shortcut: N key")
    parent.button_newcell.clicked.connect(parent.ClickNewCell)
    parent.button_newcell.setStatusTip('Use left clicks to produce a polygon with a new cell ID value.')
    
    # NEXT FRAME (TIME AXIS)
    parent.button_nextframe.toggle()
    parent.button_nextframe.pressed.connect(parent.ChangeNextFrame)
    parent.button_nextframe.setToolTip("Shortcut: right arrow (->)")
    parent.button_nextframe.setMaximumWidth(150)
    parent.button_nextframe.setShortcut(Qt.Key.Key_Right)


    
    # if there is only one picture than this button is disabled
    if np.all(parent.nextframe == 0):
        parent.button_nextframe.setEnabled(False)    
    
    # PREVIOUS FRAME (TIME AXIS)
    parent.button_previousframe.setEnabled(False)
    parent.button_previousframe.toggle()
    parent.button_previousframe.pressed.connect(parent.ChangePreviousFrame)
    parent.button_previousframe.setToolTip("Shortcut: left arrow (<-)")
    parent.button_previousframe.setMaximumWidth(150)
    parent.button_previousframe.setShortcut(Qt.Key.Key_Left)
    parent.button_previousframe.move(100,100)
        
    # ZOOM
    parent.button_zoom = QPushButton()
    parent.button_zoom.clicked.connect(parent.ZoomTlbar)
    parent.button_zoom.setIcon(QtGui.QIcon(str(path_icons) + '/ZoomIcon.png'))
    parent.button_zoom.setMaximumWidth(30)
    parent.button_zoom.setMaximumHeight(30)
    parent.button_zoom.setStyleSheet("QPushButton:hover { background-color: blue }" )
    parent.button_zoom.setCheckable(True)
    parent.button_zoom.setShortcut("Z")
    parent.button_zoom.setToolTip("Shortcut: Z")
    parent.button_zoom.setStatusTip('Zoom (draw rectangle with right mouse button to zoom out)')
    parent.buttonlist.append(parent.button_zoom)
    
    # HOME
    parent.button_home = QPushButton()
    parent.button_home.clicked.connect(parent.HomeTlbar)
    parent.button_home.setIcon(QtGui.QIcon(str(path_icons) + '/HomeIcon.png'))
    parent.button_home.setMaximumWidth(30)
    parent.button_home.setMaximumHeight(30)
    parent.button_home.setStyleSheet("QPushButton:hover { background-color: blue }" )
    parent.button_home.setShortcut("H")
    parent.button_home.setToolTip("Shortcut: H")
    parent.button_home.setStatusTip('Reset zoom')
    parent.buttonlist.append(parent.button_home)
    
    # PREVIOUS SCALE (ZOOM SCALE)
    parent.button_back = QPushButton()
    parent.button_back.clicked.connect(parent.BackTlbar)
    parent.button_back.setIcon(QtGui.QIcon(str(path_icons) +'/LeftArrowIcon.png'))
    parent.button_back.setMaximumWidth(30)
    parent.button_back.setMaximumHeight(30)
    parent.button_back.setStyleSheet("QPushButton:hover { background-color: blue }" )
    parent.button_back.setStatusTip('Go back to previous view (undo).')
    parent.buttonlist.append(parent.button_back)
    
    # NEXT SCALE (ZOOM SCALE)
    parent.button_forward = QPushButton()
    parent.button_forward.clicked.connect(parent.ForwardTlbar)
    parent.button_forward.setIcon(QtGui.QIcon(str(path_icons) + '/RightArrowIcon.png'))
    parent.button_forward.setMaximumWidth(30)
    parent.button_forward.setMaximumHeight(30)
    parent.button_forward.setStyleSheet("QPushButton:hover { background-color: blue }")
    parent.button_forward.setStatusTip('Go to next view (redo).')
    parent.buttonlist.append(parent.button_forward)
    
    # PAN 
    parent.button_pan = QPushButton()
    parent.button_pan.clicked.connect(parent.PanTlbar)
    parent.button_pan.setIcon(QtGui.QIcon(str(path_icons) + '/MoveArrowsIcon.png'))
    parent.button_pan.setMaximumWidth(30)
    parent.button_pan.setMaximumHeight(30)
    parent.button_pan.setStyleSheet("QPushButton:hover { background-color: blue }")
    parent.button_pan.setCheckable(True)
    parent.button_pan.setShortcut("P")
    parent.button_pan.setToolTip("Shortcut: P")
    parent.button_pan.setStatusTip('Pan')
    parent.buttonlist.append(parent.button_pan)
    
    # BRUSH 
    parent.button_drawmouse.setCheckable(True)
    parent.button_drawmouse.clicked.connect(parent.MouseDraw)
    parent.button_drawmouse.setToolTip("Shortcut: B")
    parent.button_drawmouse.setShortcut("B")
    parent.button_drawmouse.setMaximumWidth(150)
    parent.button_drawmouse.setStatusTip('Right click to select a cell ID value, then draw with left mouse button.')
    
    # ERASER
    parent.button_eraser.setCheckable(True)
    parent.button_eraser.clicked.connect(parent.MouseDraw)
    parent.button_eraser.setToolTip("Shortcut: E")
    parent.button_eraser.setShortcut("E")
    parent.button_eraser.setMaximumWidth(150)
    parent.button_eraser.setStatusTip('Drag to set values to background (=0).')
    
    # BRUSHSIZE
    parent.spinbox_brushsize.setMinimum(1)
    parent.spinbox_brushsize.setMaximumWidth(100)
    parent.spinbox_brushsize.setValue(3)
    
    # EXCHANGE CELL VALUES
    parent.button_exval.toggle()
    parent.button_exval.setEnabled(True)
    parent.button_exval.clicked.connect(parent.DialogBoxECV)
    parent.button_exval.setMaximumWidth(150)
    parent.button_exval.setStatusTip('Start dialog box for exchanging ID values between two cells.')

    # CHANGE CELL VALUE
    parent.button_changecellvalue.setCheckable(True)
    parent.button_changecellvalue.setEnabled(True)
    parent.button_changecellvalue.clicked.connect(parent.ChangeOneValue)
    parent.button_changecellvalue.setMaximumWidth(150)
    parent.button_changecellvalue.setShortcut("C")
    parent.button_changecellvalue.setToolTip("Shortcut: C")
#    parent.button_changecellvalue.setStatusTip('')
    parent.button_changecellvalue.setStatusTip('Change ID value of one cell. Use left click to select one cell and enter a new ID value.')

    # SPLIT CELLS
    parent.button_split.toggle()
    parent.button_split.setCheckable(True)
    parent.button_split.setEnabled(True)
    parent.button_split.clicked.connect(parent.SelectSplitCell)
    parent.button_split.setMaximumWidth(150)
    parent.button_split.setShortcut("X")
    parent.button_split.setToolTip("Shortcut: X")
    parent.button_split.setStatusTip('Split a selected cell into two. First left-click to select the cell, then draw a polygon around an area to split it off.')

    # MERGE CELLS
    parent.button_mergewithneighbors.toggle()
    parent.button_mergewithneighbors.setCheckable(True)
    parent.button_mergewithneighbors.setEnabled(True)
    parent.button_mergewithneighbors.clicked.connect(parent.MergeWithNeighbors)
    parent.button_mergewithneighbors.setMaximumWidth(150)
    parent.button_mergewithneighbors.setShortcut("F")
    parent.button_mergewithneighbors.setToolTip("Shortcut: F")
    parent.button_mergewithneighbors.setStatusTip('Merge a cell with its immediate neighbors. Left-click to select the cell or right-click to abort.')

    # MAKE THE CELL Correspondence
    parent.button_cellcorrespondence.setEnabled(False)
    parent.button_cellcorrespondence.setCheckable(True)
    parent.button_cellcorrespondence.clicked.connect(parent.BatchCellCorresp)
    parent.button_cellcorrespondence.setMaximumWidth(150)
    parent.button_cellcorrespondence.setStatusTip('Retrack cell ID numbers for next frame(s) or selected next frames based on this frame.')
        
    # EXTRACT FLUORESCENCE IN DIFFERENT CHANNELS    
    parent.button_extractfluorescence.setEnabled(False)
    parent.button_extractfluorescence.toggle()
    parent.button_extractfluorescence.clicked.connect(parent.ButtonFluo)
    parent.button_extractfluorescence.setMaximumWidth(150)
    parent.button_extractfluorescence.setStatusTip('Start the dialog box for exporting cell geometries and fluorescence for different channels.')
    
    # SHOW THE VALUES OF THE CELLS
    parent.button_showval.stateChanged.connect(parent.m.OnShowCellID)
    parent.button_showval.setShortcut('S')
    parent.button_showval.setToolTip("Shortcut: S")
        
    # HIDE/SHOW THE MASK
    parent.button_hidemask.stateChanged.connect(parent.m.HideMask)    
    parent.button_hidemask.setShortcut('M')
    parent.button_hidemask.setToolTip("Shortcut: M")
    
    # CHANGE TIME INDEX
    parent.button_timeindex = QLineEdit()
    parent.button_timeindex.setPlaceholderText('Time index 0-{}'.format(parent.reader.sizet-1))
    parent.button_timeindex.setValidator(QtGui.QIntValidator(0,int(parent.reader.sizet-1)))
    parent.button_timeindex.returnPressed.connect(parent.ChangeTimeFrame)
    parent.button_timeindex.setMaximumWidth(150)
    parent.buttonlist.append(parent.button_timeindex)
    
    # FIELDS OF VIEW BUTTON
    parent.button_fov = QComboBox()
    list_fov = []
    for i in range(0, parent.reader.Npos):
        list_fov.append("Field of View " + str(i+1))
    parent.button_fov.addItems(list_fov)
    parent.button_fov.setMaximumWidth(150)
    parent.buttonlist.append(parent.button_fov)
    parent.button_fov.activated.connect(parent.SelectFov)
            
    # CHANGE CHANNEL BUTTON
    parent.button_channel = QComboBox()
    parent.button_channel.addItems(parent.reader.channel_names)
    parent.button_channel.setMaximumWidth(150)
    parent.buttonlist.append(parent.button_channel)
    parent.button_channel.activated.connect(parent.SelectChannel)
        
    # NEURAL NETWORK BUTTON
    parent.button_cnn.setCheckable(True)
    parent.button_cnn.pressed.connect(parent.LaunchBatchPrediction)
    parent.button_cnn.setStatusTip("Start the dialog box for segmenting images.")
    parent.button_cnn.setMaximumWidth(150)
    parent.EnableCNNButtons()
        




# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 17:38:58 2019
"""

from PyQt6.QtWidgets import (QPushButton, QDialog, QDialogButtonBox, 
                             QLineEdit, QFormLayout, QMessageBox, 
                             QFileDialog, QLabel)
import os

class FileBrowser(QDialog):

    def __init__(self, *args, **kwargs):
        super(FileBrowser, self).__init__(*args, **kwargs)
        
        self.setWindowTitle("Open Files")
        self.setGeometry(100,100, 800,200)
        
        self.button_opennd2 = QPushButton('Open image file')
        self.button_opennd2.setEnabled(True)
        self.button_opennd2.clicked.connect(self.getnd2path)
        self.button_opennd2.setToolTip("Browse for an image file")
        self.button_opennd2.setMaximumWidth(150)
        
        self.button_openfolder = QPushButton('Open image folder')
        self.button_openfolder.setEnabled(True)
        self.button_openfolder.clicked.connect(self.getfolder)
        self.button_openfolder.setToolTip("Browse for folder with images")
        self.button_openfolder.setMaximumWidth(150)
        
        self.button_openhdf = QPushButton('Open mask file')
        self.button_openhdf.setEnabled(True)
        self.button_openhdf.clicked.connect(self.gethdfpath)
        self.button_openhdf.setToolTip("Browse for a mask file")
        self.button_openhdf.setMaximumWidth(150)
        
        self.newhdfentry = QLineEdit()
        self.newhdfentry.setText("_new_mask.h5")

        self.nd2name = ''
        self.hdfname = ''
        flo = QFormLayout()
        
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


        self.labelnd2 = QLabel()
        self.labelnd2.setText('No image file (.nd2, .tif, .tiff) selected')
        
        self.labelhdf = QLabel()
        self.labelhdf.setText('No mask file (.h5, .tif, .tiff) selected')
        
        self.labelfolder = QLabel()
        self.labelfolder.setText('No folder selected')
        
        flo.addRow(self.labelnd2, self.button_opennd2)
        flo.addRow(self.labelfolder, self.button_openfolder)
        flo.addRow(self.labelhdf, self.button_openhdf)
        flo.addRow('If no hdf file already exists, give a name to create a new file', self.newhdfentry)
        
        flo.addWidget(self.buttonBox)
       
        self.setLayout(flo)
        
        
        

    def getnd2path(self):
        self.nd2name,_ = QFileDialog.getOpenFileName(self, 'Open image file','', 
            'Image files (*.nd2 *.tif *.tiff *.jpg *.jpeg *.png *.bmp '
                          '*.pbm *.pgm *.ppm *.pxm *.pnm *.jp2 '
                          '*.TIF *.TIFF *.JPG *.JPEG *.PNG *.BMP '
                          '*.PBM *.PGM *.PPM *.PXM *.PNM *.JP2)')
        self.nd2name = self.nd2name
        if self.nd2name != '':
            path , ext = os.path.splitext(self.nd2name)
            self.labelnd2.setText(str(self.nd2name))
            self.labelfolder.setText('')
            self.newhdfname = path +'_new_mask.h5'
            self.newhdfentry.setText(self.newhdfname)
            self.check_new_hdfpath()

      
    def gethdfpath(self):
        self.hdfname,_ = QFileDialog.getOpenFileName(self,'Open mask file','', 'Mask files (*.h5 *.tif *.tiff)')
        if self.hdfname != '':
            self.check_hdfpath()
            self.labelhdf.setText(self.hdfname)
            self.newhdfentry.setText("")
            self.newhdfname = ''
        
    def getfolder(self):
        folder = QFileDialog.getExistingDirectory(self, ("Select Image Folder"))
        if folder != '':
            self.nd2name = folder
            self.labelfolder.setText(self.nd2name)
            self.labelnd2.setText('')
            self.newhdfname = os.path.join(str(self.nd2name), '_new_mask.h5')
            self.newhdfentry.setText(self.newhdfname)
            self.check_new_hdfpath()
        else: # if canceled
            self.labelnd2.setText(self.nd2name) # keep old path

            
    def check_hdfpath(self):
        """Checks if hdf path already exists when loading tiff, to avoid 
        data loss"""
        path, ext = os.path.splitext(self.hdfname)
        if ext=='.tiff' or ext=='.tif' or ext=='.TIFF' or ext=='.TIF':
            if os.path.isfile(path+'.h5'):
                msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Warning',
                                 'A .h5 file with the same name as the loaded '
                                 'tif exists already and will be overwritten.'
                                 ' Rename either the tif or the h5 file to '
                                 'avoid data loss.', parent=self)
                msg_box.exec()

    def check_new_hdfpath(self):
        """Checks if hdf path already exists when creating new hdf, to avoid 
        data loss"""
        if self.newhdfentry.text() != '':
            if os.path.isfile(self.newhdfname):
                msg_box = QMessageBox(QMessageBox.Icon.Critical, 'Warning',
                                 'A .h5 file with the same name as the new '
                                 'hdf exists already and will be overwritten.'
                                 ' Rename either the new hdf or the existing '
                                 'hdf file to avoid data loss.', parent=self)
                msg_box.exec()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This file handles the PlotCanvas of the GUI - aka the area where the graphics 
are shown. 
"""
import numpy as np
from skimage import morphology as morph
from skimage import draw
import logging
import os
logging.basicConfig(
    format='%(asctime)s %(levelname)s %(funcName)s: %(message)s',
    level=os.environ.get("LOGLEVEL", "WARNING")
)
log = logging.getLogger(__name__)


# Import everything for the Graphical User Interface from the PyQt6 library.
from PyQt6.QtWidgets import QSizePolicy

#Import from matplotlib to use it to display the pictures and masks.
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

import matplotlib.pyplot as plt

from matplotlib import cm
from matplotlib.colors import ListedColormap
#from matplotlib.path import Path

from scipy import ndimage

from PIL import Image, ImageDraw



class PlotCanvas(FigureCanvas):
    def __init__(self, parent=None):
        """this class defines the canvas. It initializes a figure, which is then
        used to plot our data using imshow.
        """
        # define three subplots corresponding to the previous, current and next
        # time index.
        fig, (self.ax2, self.ax, self.ax3) = plt.subplots(1,3, sharex = True, sharey = True)
        
        fig.subplots_adjust(bottom=0, top=1, left=0, right=1, wspace = 0.05, hspace = 0.05)
        FigureCanvas.__init__(self, fig)
        self.setParent(parent)
        self.parent = parent
        
        # this is some mambo jambo.
        FigureCanvas.setSizePolicy(self, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        FigureCanvas.updateGeometry(self)
        
        # the self.currpicture attribute takes the original data and will then 
        # contain the updates drawn by the user.
        self.currpicture = parent.currentframe
        self.prevpicture = parent.previousframe
        self.nextpicture = parent.nextframe
        self.plotmask = parent.mask_curr
        self.prevplotmask = parent.mask_previous
        self.nextplotmask = parent.mask_next
        self.tempmask = self.plotmask.copy()
        self.tempplotmask = self.plotmask.copy()
        
        # this line is just here to not attribute a zero value to the plot
        # because if so, then it does not update the plot and it stays blank.
        self.prevpicture = self.currpicture.copy()
        
        # Initialize Plots
        self.currplot, self.currmask = self.plot(self.currpicture, self.plotmask, self.ax)
        
        self.previousplot, self.previousmask = self.plot(self.prevpicture, self.prevplotmask, self.ax2)
        self.prevpicture = np.zeros([parent.reader.sizey, parent.reader.sizex], dtype = np.uint16)
        self.prevplotmask = np.zeros([parent.reader.sizey, parent.reader.sizex], dtype =np.uint16)
        
        self.nextplot, self.nextmask = self.plot(self.nextpicture, self.nextplotmask, self.ax3)
        self.previousplot.set_data(self.prevpicture)
        self.previousmask.set_data((self.prevplotmask%10+1)*(self.prevplotmask != 0))

        self.ax2.draw_artist(self.previousplot)
        self.ax2.draw_artist(self.previousmask)
        self.update()
        self.flush_events()
        
        # Set title labels
        self.titlecurr = self.ax.set_title('Time index {}'.format(parent.Tindex))
        self.titleprev = self.ax2.set_title('No frame {}'.format(''))
        self.titlenext = self.ax3.set_title('Next time index {}'.format(parent.Tindex+1))
        
        # these variables are just set to test the states of the buttons
        # (button turned on or  off, etc..) of the buttons in the methods 
        # used in this class.
        self.button_showval_check = parent.button_showval
        self.button_newcell_check = parent.button_newcell
        self.button_add_region_check = parent.button_add_region
        self.button_drawmouse_check = parent.button_drawmouse
        self.button_eraser_check = parent.button_eraser
        self.button_hidemask_check = parent.button_hidemask
        
        # It will plot for the first time and return the imshow function
        self.currmask.set_clim(0, 10)
        self.previousmask.set_clim(0,10)
        self.nextmask.set_clim(0,10)
        
        # This attribute is a list which stores all the clicks of the mouse.
        self.storemouseclicks = []
        
        # This attribute is used to store the square where the mouse has been
        # in order than to draw lines (Paintbrush function)
        self.storebrushclicks = [False,False]
        
        # self.cellval is the variable which sets the value to the pixel
        # whenever something is drawn.
        self.cellval = 0
        
        # These are lists storing all the annotations which are used to
        # show the values of the cells on the plots.
        self.ann_list = []
        self.ann_list_prev = []
        self.ann_list_next = []
    
        self.cid = self.mpl_connect('motion_notify_event', self.on_motion)


    def on_motion(self, event):
        if event.inaxes == self.ax:
            x, y = event.xdata, event.ydata
            # Update the label texts with the current mouse position
            self.parent.WriteStatusBar(f'X: {x}, Y: {y}')

    def ExchangeCellValue(self, val1, val2):
        """Swaps the values of the cell between two clusters each representing
        one cell. This method is called after the user has entered 
        values in the ExchangeCellValues window.
        """
        if (val1 in self.plotmask) and (val2 in self.plotmask):
            indices = np.where(self.plotmask == val1)
            self.plotmask[self.plotmask == val2] = val1
            for i in range(0,len(indices[0])):
                self.plotmask[indices[0][i], indices[1][i]] = val2
            self.updatedata()
        else:
            raise ValueError('Cell value does not exist.') 
        
        
    def ReleaseClick(self, event):
        """This method is called from the brush button when the mouse is 
        released such that the last coordinate saved when something is drawn
        is set to zero. Because otherwise, if the user starts drawing somewhere
        else, than a straight line is draw between the last point of the
        previous mouse drawing/dragging and the new one which then starts.
        """
        if self.ax == event.inaxes:
            self.storebrushclicks = [False, False]
            self.ShowCellNumbers(type = 'current')

        
    def OneClick(self, event, radius=3):
        """This method is called when the Brush button is activated. And
        sets the value of self.cellval if the click is a right click, or draws
        a square if the click is a left click. (so if the user does just left
        click but does not drag, there will be only a square which is drawn )
        """
        selem = morph.disk(radius-1)
        
        # Right click selects cell
        if (event.button == 3 
            and (event.xdata != None and event.ydata != None) 
            and (not self.button_eraser_check.isChecked()) 
            and self.ax == event.inaxes):
            tempx = int(event.xdata)
            tempy = int(event.ydata)
            self.cellval = self.plotmask[tempy, tempx]
            self.storebrushclicks = [False, False]
            
        # Left click draws square
        elif (event.button == 1 and 
              (event.xdata != None and event.ydata != None) 
              and self.ax == event.inaxes):
            tempx = int(event.xdata)
            tempy = int(event.ydata)
            
            to_change = np.zeros(self.plotmask.shape, dtype=bool)
            to_change[tempy, tempx] = True
            to_change = morph.dilation(to_change, selem)
            
            self.plotmask[to_change] = self.cellval
            self.storebrushclicks = [tempx,tempy]
            self.updatedata()
            
        else:
            return
        
        
    def multiple_click(self, event, call_after, radius=3):
        """Function to keep track of multiple left clicks, confirmed with 
        a right click. After right click, the function call_after is called"""        
        
        if (event.button == 1  # left click
            and (event.xdata != None and event.ydata != None) 
            and self.ax == event.inaxes):
            
            newx = int(event.xdata)
            newy = int(event.ydata)
            self.storemouseclicks.append((newx, newy))
            
            self.updateplot(newx, newy, False)
        
        elif (event.button == 3): # right click
            call_after()

            
    
    def PaintBrush(self, event, radius=3):
        """PantBrush is the method to paint using a "brush" and it is based
        on the mouse event in matplotlib "motion notify event". However it can 
        not record every pixel that the mouse has hovered over (it is too fast).
        So, in order to not only draw points (happens when the mouse is dragged
        too quickly), these points are interpolated here with lines.
        """
        selem = morph.disk(radius-1)
        if (event.button == 1 
            and (event.xdata != None and event.ydata != None) 
            and self.ax == event.inaxes):
            newx = int(event.xdata)
            newy = int(event.ydata)
            # when a new cell value is set, there is no point to interpolate, to
            # draw a line between the points. 
            if self.storebrushclicks[0] == False :
                self.storebrushclicks = [newx,newy]
                
            else:
                oldx, oldy = self.storebrushclicks
                
                rr, cc, _ = draw.line_aa(newy, newx, oldy, oldx)
                to_change = np.zeros(self.plotmask.shape, dtype=bool)
                to_change[rr,cc] = True
                to_change = morph.dilation(to_change, selem)
                self.plotmask[to_change] = self.cellval
                
            self.storebrushclicks = [newx, newy]
            self.updatedata()
            
            
    def MouseClick(self,event):
        """This function is called whenever the add region or the new cell
        buttons are active and the user clicks on the plot. For each 
        click on the plot, it records the coordinate of the click and stores
        it. When the user deactivate the new cell or add region button, 
        all the coordinates are given to the DrawRegion function (if they 
        do not all lie on the same line) and out of the coordinates, it makes
        a polygon. And then draws inside of this polygon by setting the pixels
        to the self.cellval value.
        """
        # button == 1 corresponds to the left click. 
        if (event.button == 1 
            and (event.xdata != None and event.ydata != None) 
            and self.ax == event.inaxes):
            
            # extract the coordinate of the click inside of the matplotlib figure
            # and then takes the integer part
            newx = int(event.xdata)
            newy = int(event.ydata)
            
            # stores the coordinates of the click
            self.storemouseclicks.append((newx, newy))
            
            # draws in the figure a small square (4x4 pixels) to
            # visualize where the user has clicked
            self.updateplot(newx, newy)
                

    def DefineColormap(self, Ncolors):
        """Define a new colormap by assigning 10 values of the jet colormap
            such that there are only colors for the values 0-10 and the values >10
            will be treated with a modulo operation (updatedata function)
        """
        jet = cm.get_cmap('jet', Ncolors)
        colors = []
        for i in range(0,Ncolors):
            if i==0 : 
                # set background transparency to 0
                temp = list(jet(i))
                temp[3]= 0.0
                colors.append(tuple(temp))
                
            else:
                colors.append(jet(i))
                
        colormap = ListedColormap(colors)
        return colormap
           
    
    def plot(self, picture, mask, ax):
        """this function is called for the first time when all the subplots
        are drawn.
        """
        # Define a new colormap with 20 colors.
        newcmp = self.DefineColormap(21)
        ax.axis("off")

        self.draw()
        return (ax.imshow(picture, interpolation= 'None', 
                        origin = 'lower', cmap = 'gray_r'), 
                ax.imshow((mask%10+1)*(mask != 0), origin = 'lower', 
                        interpolation = 'None', alpha = 0.2, cmap = newcmp))
   
    
    def UpdatePlots(self, type = "all"):
        """
        Updates plots, handles mask and cell numbers.
        """
        
        # Plot images
        log.debug('call UpdatePlot')
        self.currplot.set_data(self.currpicture)
        self.currplot.set_clim(np.amin(self.currpicture), np.amax(self.currpicture))
        self.ax.draw_artist(self.currplot)
        
        self.previousplot.set_data(self.prevpicture)
        self.previousplot.set_clim(np.amin(self.prevpicture), np.amax(self.prevpicture))
        self.ax2.draw_artist(self.previousplot)
            
        self.nextplot.set_data(self.nextpicture)
        self.nextplot.set_clim(np.amin(self.nextpicture), np.amax(self.nextpicture))
        self.ax3.draw_artist(self.nextplot)
        
        # Plot masks
        if not self.button_hidemask_check.isChecked():
            self.currmask.set_data((self.plotmask%10+1)*(self.plotmask!=0))
            self.previousmask.set_data((self.prevplotmask%10+1)*(self.prevplotmask != 0))
            self.nextmask.set_data((self.nextplotmask % 10 +1 )*(self.nextplotmask != 0))
            
        else:
            self.currmask.set_data(np.zeros(self.plotmask.shape))
            self.previousmask.set_data(np.zeros(self.plotmask.shape))
            self.nextmask.set_data(np.zeros(self.plotmask.shape))
        
        self.ShowCellNumbers(type=type)
        self.draw()
        self.update()
        self.flush_events()
                
        
    def updatedata(self, flag=True):
        """
        In order to just display the cells so regions with value > 0
        and also to assign to each of the cell values one color,
        the modulo 10 of the value is take and we add 1, to distinguish
        the values of 10,20,30,... from the background (although the bckgrnd
        gets with the addition the value 1) and the result of the 
        modulo is multiplied with a matrix containing a False value for the 
        background coordinates, setting the background to 0 again.
        """
        if flag:
            self.currmask.set_data((self.plotmask%10+1)*(self.plotmask!=0))
        else:
            self.currmask.set_data((self.tempmask%10+1)*(self.tempmask!=0))
        log.debug('updatedata with flag {}'.format(flag))
        # show the updates by redrawing the array using draw_artist, it is faster 
        # to use as it only redraws the array itself, and not everything else.
        self.ax.draw_artist(self.currplot)
        self.ax.draw_artist(self.currmask)
        self.ShowCellNumbers(type='current' if flag else 'None')
        self.update()
        # self.flush_events()
              
        
    def HideMask(self):
        self.UpdatePlots()
        
                 
    def _getCellCenters(self, plotmask):
        """Get approximate locations for cell centers"""
        log.debug('start')
        vals = np.unique(plotmask).astype(int)
        vals = np.delete(vals, np.where(vals == 0))
        centers = np.array(ndimage.measurements.center_of_mass(plotmask, labels=plotmask, index=vals))
        xtemp = np.round(centers[:, 1]).astype(int)
        ytemp = np.round(centers[:, 0]).astype(int)
        log.debug('finish')
        return vals, xtemp, ytemp

    def OnShowCellID(self):
        """This function is only called when we activate the show cell IDs checkbox."""
        self.UpdatePlots(type='all')

    def ShowCellNumbers(self, type='all'):
        """
        Checks whether to show cell numbers, and does so if button is 
        checked
        This function can apply to current frame or to all frames
        """
        log.debug("show cell numbers called with type {}".format(type))
        if self.button_showval_check.isChecked():
            if(type == 'current'):
                self.ShowCellNumbersCurr()
            elif(type == 'all'):
                self.ShowCellNumbersCurr()
                self.ShowCellNumbersNext()
                self.ShowCellNumbersPrev()
            self.draw()
        else:
            self.clearAnnLists()
        
    
    def ShowCellNumbersCurr(self):
        """This function is called to display the cell values and computes the cebter of each cell and
        gives the coordinate where the number will be 
        displayed. The number to be displayed is just given by the value
        in the mask of the cell.
        This function is just used for the current time subplot.
        """         
        
        for i,a in enumerate(self.ann_list):
            a.remove()
        self.ann_list[:] = []
        # check if the mask is not empty
        if np.sum(self.plotmask)==0 :
            return
            
        vals, xtemp, ytemp = self._getCellCenters(self.plotmask)
        if xtemp.any():
            for i in range(0,len(xtemp)):
                ann = self.ax.annotate(str(int(vals[i])), (xtemp[i], ytemp[i]),
                                        ha='center', va='center')
                self.ann_list.append(ann)
    #  self.draw()
                     
             
    def ShowCellNumbersPrev(self):
        """This function is called to display the cell values and computes the cebter of each cell and
        gives the coordinate where the number will be 
        displayed. The number to be displayed is just given by the value
        in the mask of the cell.
        This function is just used for the previous time subplot and check if it is the first frame or not.
        """
        
        for i,a in enumerate(self.ann_list_prev):
            a.remove()
        self.ann_list_prev[:] = []
        
        if(self.parent.Tindex != 0):
            if np.sum(self.prevplotmask)==0 :
                return
            
            vals, xtemp, ytemp = self._getCellCenters(self.prevplotmask)
            if xtemp.any():
                for i in range(0,len(xtemp)):
                    ann = self.ax2.annotate(str(vals[i]), (xtemp[i], ytemp[i]),
                                            ha='center', va='center')
                    self.ann_list_prev.append(ann)
        #  self.draw()
        else:
            self.ann_list_prev = []
             
             
    def ShowCellNumbersNext(self):
        """This function is called to display the cell values and computes the cebter of each cell and
        gives the coordinate where the number will be 
        displayed. The number to be displayed is just given by the value
        in the mask of the cell.
        This function is just used for the next time subplot and check if it is the first frame or not.
        """
        for i,a in enumerate(self.ann_list_next):
            a.remove()
        self.ann_list_next[:] = []
        if(self.parent.Tindex != self.parent.reader.sizet-1):
            if np.sum(self.nextplotmask)==0 :
                return           
            vals, xtemp, ytemp = self._getCellCenters(self.nextplotmask)
                
            if xtemp.any():
                for i in range(0,len(xtemp)):
                    ann = self.ax3.annotate(str(vals[i]), (xtemp[i], ytemp[i]),
                                        ha='center', va='center')
                    self.ann_list_next.append(ann)
        #  self.draw()
        else: 
            self.ann_list_next = []
        
        
    def clearAnnLists(self):
        for ann in self.ann_list:
            ann.remove()
        self.ann_list = []
        for ann in self.ann_list_prev:
            ann.remove()
        self.ann_list_prev = []
        for ann in self.ann_list_next:
            ann.remove()
        self.ann_list_next = []
        
        
    def updateplot(self, posx, posy, first_is_cell=True):
        """
        it updates the plot once the user clicks on the plot and draws a 4x4 pixel dot
        at the coordinate of the click 
        """        
        if first_is_cell:
            # remove the first coordinate as it should only coorespond 
            # to the value that the user wants to attribute to the drawn region   
            xtemp, ytemp = self.storemouseclicks[0]
    
            # here we initialize the value attributed to the pixels.
            # it means that the first click selects the value that will be attributed to
            # the pixels inside the polygon (drawn by the following mouse clicks of the user)
            self.cellval = self.plotmask[ytemp, xtemp]
            
        else:
            self.cellval=1
          
        # drawing the 2x2 square ot of the mouse click
        if ((self.button_newcell_check.isChecked() or self.button_drawmouse_check.isChecked()) 
            and self.cellval == 0):
            self.tempmask[posy:posy+2, posx:posx+2] = 9
        else:
            self.tempmask[posy:posy+2,posx:posx+2] = self.cellval

        # plot the mouseclick
        self.updatedata(False)
          

    def DrawRegion(self, flag):
        """
        this method is used to draw either a new cell (flag = true) or to add a region to 
        an existing cell (flag = false). The flag will just be used to set the
        value of pixels (= self.cellval) in the drawn region. 
        If flag = true, then the value will be the maximal value plus 1. Such 
        that it attributes a new value to the new cell.
        If flag = false, then it will use the value of the first click to set
        the value of the pixels in the new added region. 
        """
        # here the values that have been changed to mark the mouse clicks are 
        # restored such that they don't appear when the region/new cell is 
        # drawn.        
        if flag:
            # if new cell is added, it sets the value of the drawn pixels to a new value
            # corresponding to the new cell
            self.cellval = np.amax(self.plotmask) + 1
            
        else:
            # The first value is taken out as it is just used to set the value
            # to the new region.
            self.storemouseclicks.pop(0)
        
        if len(self.storemouseclicks) <= 2:
            # if only two points or less have been click, it cannot make a area
            # so it justs discards these values and returns. 
            self.storemouseclicks = list(self.storemouseclicks)
            self.storemouseclicks.clear()
            self.updatedata(True)
            return
        
        else:
            # Draw polygon, fill it with cell values
            nx, ny = self.plotmask.shape
            img = Image.new('L', (ny, nx), 0)
            ImageDraw.Draw(img).polygon(self.storemouseclicks, outline=1, fill=1)
            polygon = np.array(img).astype(bool)
            self.plotmask[polygon] = self.cellval
            self.updatedata()
            
        # empty the lists ready for the next region to be drawn.
        self.storemouseclicks = []
        